import os
import astropy.io.fits as fits
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import scipy.constants as sc

from .parameters import Params, find_parameter_file
from .utils import *

class Dust_model:

    def __init__(self, dir=None, **kwargs):
        # Correct path if needed
        dir = os.path.normpath(os.path.expanduser(dir))
        self.dir = dir

        # Search for parameter file
        para_file = find_parameter_file(dir)

        # Read parameter file
        self.P = Params(para_file)

        # Read model results
        self._read(**kwargs)


    def _read(self):

        try:
            self.wl = fits.getdata(self.dir+"/lambda.fits.gz")
            self.kappa = fits.getdata(self.dir+"/kappa.fits.gz")
            self.albedo = fits.getdata(self.dir+"/albedo.fits.gz")

            self.kappa_abs = self.kappa * (1-self.albedo)
            self.kappa_sca = self.kappa * self.albedo

            self.phase_function = fits.getdata(self.dir+"/phase_function.fits.gz")
            try:
                self.polarisability = fits.getdata(self.dir+"/polarizability.fits.gz")
            except:
                print("Missing polarisability")


        except OSError:
            print('cannot open dust model in', self.dir)

    def plot_kappa(self,abs=True,scatt=True, linewidth=1, **kwargs):

        plt.loglog(self.wl,self.kappa,linewidth=linewidth, color="black")
        if abs:
            plt.loglog(self.wl,self.kappa_abs,linewidth=linewidth, color="red")
        if scatt:
            plt.loglog(self.wl,self.kappa_sca,linewidth=linewidth, color="blue")

        plt.xlabel("$\lambda$ [$\mu$m]")
        plt.ylabel("$\kappa$ [cm$^2$/g]")

    def print_kappa(self,file="opacities.txt"):
        np.savetxt(file, np.array([self.wl,self.kappa_abs,self.kappa_sca]).T ,header="# wl [micron]  kappa_abs [cm2/g] kappa_abs [cm2/g]")

    def plot_albedo(self,linewidth=1, **kwargs):

        plt.semilogx(self.wl,self.albedo,linewidth=linewidth)
        plt.xlabel("$\lambda$ [$\mu$m]")
        plt.ylabel("albedo")

    def plot_phase_function(self,k=0,linewidth=1, **kwargs):
        theta = np.arange(0,181)
        plt.semilogy(theta,self.phase_function[:,k])

    def plot_polarisability(self,k=0,linewidth=1, **kwargs):
        theta = np.arange(0,181)
        plt.plot(theta,self.polarisability[:,k])



import numpy as np
from scipy.interpolate import interp1d
from scipy import ndimage
import scipy.constants as sc
import astropy.constants as const
import astropy.units as u
import cmasher as cmr

default_cmap = cmr.arctic

sigma_to_FWHM = 2.0 * np.sqrt(2.0 * np.log(2))
FWHM_to_sigma = 1.0 / sigma_to_FWHM
arcsec = np.pi / 648000


def bin_image(im, n, func=np.sum):
    # bin an image in blocks of n x n pixels
    # return a image of size im.shape/n

    nx = im.shape[0]
    nx_new = nx // n
    x0 = (nx - nx_new * n) // 2

    ny = im.shape[1]
    ny_new = ny // n
    y0 = (ny - ny_new * n) // 2

    return np.reshape(
        np.array(
            [
                func(im[x0 + k1 * n : (k1 + 1) * n, y0 + k2 * n : (k2 + 1) * n])
                for k1 in range(nx_new)
                for k2 in range(ny_new)
            ]
        ),
        (nx_new, ny_new),
    )


def Wm2_to_Jy(nuFnu, nu):
    '''
    Convert from W.m-2 to Jy
    nu [Hz]
    '''
    return 1e26 * nuFnu / nu


def Jy_to_Wm2(Fnu, nu):
    '''
    Convert from Jy to W.m-2
    nu [Hz]
    '''
    return 1e-26 * Fnu * nu


def Jybeam_to_Tb(Fnu, nu, bmaj, bmin):
    '''
     Convert Flux density in Jy/beam to brightness temperature [K]
     Flux [Jy]
     nu [Hz]
     bmaj, bmin in [arcsec]

     T [K]
    '''
    beam_area = bmin * bmaj * arcsec ** 2 * np.pi / (4.0 * np.log(2.0))
    exp_m1 = 1e26 * beam_area * 2.0 * sc.h / sc.c ** 2 * nu ** 3 / Fnu
    hnu_kT = np.log1p(np.maximum(exp_m1, 1e-10))

    Tb = sc.h * nu / (hnu_kT * sc.k)

    return Tb


def Jy_to_Tb(Fnu, nu, pixelscale):
    '''
     Convert Flux density in Jy/pixel to brightness temperature [K]
     Flux [Jy]
     nu [Hz]
     bmaj, bmin in [arcsec]

     T [K]
    '''
    pixel_area = (pixelscale * arcsec) ** 2
    exp_m1 = 1e26 * pixel_area * 2.0 * sc.h / sc.c ** 2 * nu ** 3 / Fnu
    hnu_kT = np.log1p(np.maximum(exp_m1, 1e-10))

    Tb = sc.h * nu / (hnu_kT * sc.k)

    return Tb


def Wm2_to_Tb(nuFnu, nu, pixelscale):
    """Convert flux converted from Wm2/pixel to K using full Planck law.
        Convert Flux density in Jy/beam to brightness temperature [K]
        Flux [W.m-2/pixel]
        nu [Hz]
        bmaj, bmin, pixelscale in [arcsec]
        """
    pixel_area = (pixelscale * arcsec) ** 2
    exp_m1 = pixel_area * 2.0 * sc.h * nu ** 4 / (sc.c ** 2 * nuFnu)
    hnu_kT = np.log1p(exp_m1)

    Tb = sc.h * nu / (sc.k * hnu_kT)

    return Tb


# -- Functions to deal the synthesized beam.
def _beam_area(self):
    """Beam area in arcsec^2"""
    return np.pi * self.bmaj * self.bmin / (4.0 * np.log(2.0))

def _beam_area_str(self):
    """Beam area in steradian^2"""
    return self._beam_area() * arcsec ** 2

def _pixel_area(self):
    return self.pixelscale ** 2

def _beam_area_pix(self):
    """Beam area in pix^2."""
    return self._beam_area() / self._pixel_area()


def telescope_beam(wl,D):
    """ wl and D in m, returns FWHM in arcsec"""
    return 0.989 * wl/D / 4.84814e-6


def make_cut(im, x0,y0,x1,y1,num=None,plot=False):
    """
    Make a cut in image 'im' along a line between (x0,y0) and (x1,y1)
    x0, y0,x1,y1 are pixel coordinates
    """

    if plot:
        vmax = np.max(im)
        vmin = vmax * 1e-6
        norm = mcolors.LogNorm(vmin=vmin, vmax=vmax, clip=True)
        plt.imshow(im,origin="lower", norm=norm)
        plt.plot([x0,x1],[y0,y1])


    if num is not None:
        # Extract the values along the line, using cubic interpolation
        x, y = np.linspace(x0, x1, num), np.linspace(y0, y1, num)
        zi = ndimage.map_coordinates(im, np.vstack((y,x)))
    else:
        # Extract the values along the line at the pixel spacing
        length = int(np.hypot(x1-x0, y1-y0))
        x, y = np.linspace(x0, x1, length), np.linspace(y0, y1, length)
        zi = im[y.astype(int), x.astype(int)]

    return zi


class DustExtinction:

    import os

    __dirname__ = os.path.dirname(__file__)

    wl = []
    kext = []

    _extinction_dir = __dirname__ + "/extinction_laws"
    _filename_start = "kext_albedo_WD_MW_"
    _filename_end = "_D03.all"
    V = 5.47e-1  # V band wavelength in micron

    def __init__(self, Rv=3.1, **kwargs):
        self.filename = (
            self._extinction_dir
            + "/"
            + self._filename_start
            + str(Rv)
            + self._filename_end
        )
        self._read(**kwargs)

    def _read(self):

        with open(self.filename, 'r') as file:
            f = []
            for line in file:
                if (not line.startswith("#")) and (
                    len(line) > 1
                ):  # Skipping comments and empty lines
                    line = line.split()
                    self.wl.append(float(line[0]))
                    kpa = float(line[4])
                    albedo = float(line[1])
                    self.kext.append(kpa / (1.0 - albedo))

            # Normalize extinction in V band
            kext_interp = interp1d(self.wl, self.kext)
            kextV = kext_interp(self.V)
            self.kext /= kextV

    def redenning(self, wl, Av):
        """
        Computes extinction factor to apply for a given Av
        Flux_red = Flux * redenning

        wl in micron

        """
        kext_interp = interp1d(self.wl, self.kext)
        kext = kext_interp(wl)
        tau_V = 0.4 * np.log(10.0) * Av

        return np.exp(-tau_V * kext)


def Hill_radius():
    pass
    #d * (Mplanet/3*Mstar)**(1./3)


def splash2mcfost(anglex, angley, anglez):
    #Convert the splash angles to mcfost angles

    # Base unit vector
    x0 = [1,0,0]
    y0 = [0,1,0]
    z0 = [0,0,1]

    # Splash rotated vectors
    x = _rotate_splash_axes(x0,-anglex,-angley,-anglez)
    y = _rotate_splash_axes(y0,-anglex,-angley,-anglez)
    z = _rotate_splash_axes(z0,-anglex,-angley,-anglez)

    # MCFOST angles
    mcfost_i = np.arccos(np.dot(z,z0)) * 180./np.pi

    if abs(mcfost_i) > 1e-30:
        print("test1")
        # angle du vecteur z dans le plan (-y0,x0)
        mcfost_az = (np.arctan2(np.dot(z,x0), -np.dot(z,y0)) ) * 180./np.pi
        # angle du vecteur z0 dans le plan x_image, y_image (orientation astro + 90deg)
        mcfost_PA = -( np.arctan2(np.dot(x,z0), np.dot(y,z0)) )  * 180./np.pi
    else:
        print("test2")
        mcfost_az = 0.
        # angle du vecteur y dans le plan x0, y0
        mcfost_PA = (np.arctan2(np.dot(y,x0),np.dot(y,y0)) ) * 180./np.pi


    print("anglex =",anglex, "angley=", angley, "anglez=", anglez,"\n")
    print("Direction to oberver=",z)
    print("x-image=",x)
    print("y_image = ", y,"\n")
    print("MCFOST parameters :")
    print("inclination =", mcfost_i)
    print("azimuth =", mcfost_az)
    print("PA =", mcfost_PA)

    return [mcfost_i, mcfost_az, mcfost_PA]

def _rotate_splash(xyz, anglex, angley, anglez):
    # Defines rotations as in splash
    # This function is to rotate the data

    x = xyz[0]
    y = xyz[1]
    z = xyz[2]

    # rotate about z
    if np.abs(anglez) > 1e-30:
        r = np.sqrt(x**2+y**2)
        phi = np.arctan2(y,x)
        phi -= anglez/180*np.pi
        x = r*np.cos(phi)
        y = r*np.sin(phi)

    # rotate about y
    if np.abs(angley) > 1e-30:
        r = np.sqrt(z**2+x**2)
        phi = np.arctan2(z,x)
        phi -= angley/180*np.pi
        x = r*np.cos(phi)
        z = r*np.sin(phi)

    # rotate about x
    if np.abs(anglex) > 1e-30:
        r = np.sqrt(y**2+z**2)
        phi = np.arctan2(z,y)
        phi -= anglex/180*np.pi
        y = r*np.cos(phi)
        z = r*np.sin(phi)

    return np.array([x,y,z])


def _rotate_splash_axes(xyz, anglex, angley, anglez):
    # Defines rotations as in splash, but in reserve order
    # as we rotate the axes instead of the data

    x = xyz[0]
    y = xyz[1]
    z = xyz[2]

    # rotate about x
    if np.abs(anglex) > 1e-30:
        r = np.sqrt(y**2+z**2)
        phi = np.arctan2(z,y)
        phi -= anglex/180*np.pi
        y = r*np.cos(phi)
        z = r*np.sin(phi)

    # rotate about y
    if np.abs(angley) > 1e-30:
        r = np.sqrt(z**2+x**2)
        phi = np.arctan2(z,x)
        phi -= angley/180*np.pi
        x = r*np.cos(phi)
        z = r*np.sin(phi)

    # rotate about z
    if np.abs(anglez) > 1e-30:
        r = np.sqrt(x**2+y**2)
        phi = np.arctan2(y,x)
        phi -= anglez/180*np.pi
        x = r*np.cos(phi)
        y = r*np.sin(phi)

    return np.array([x,y,z])

def rotate_vec(u,v,angle):
    '''
      rotate a vector (u) around an axis defined by another vector (v)
      by an angle (theta) using the Rodrigues rotation formula
    '''
    k = v/np.sqrt(np.inner(v,v))
    w = np.cross(k,u)
    k_dot_u = np.inner(k,u)
    for i,uval in enumerate(u):
        u[i] = u[i]*np.cos(angle) + w[i]*np.sin(angle) + k[i]*k_dot_u*(1.-np.cos(angle))
    return u

def rotate_coords(x,y,z,inc,PA):
    '''
       rotate x,y,z coordinates into the observational plane
    '''
    k = [-np.sin(PA), np.cos(PA), 0.]
    xvec = [x,y,z]
    xrot = rotate_vec(xvec,k,inc)
    return xrot[0],xrot[1],xrot[2]

def rotate_to_obs_plane(x,y,inc,PA):
    '''
       same as rotate_coords but takes 2D x,y as arrays
    '''
    for i,xx in enumerate(x):    # this can probably be done more efficiently
        x[i],y[i],dum = rotate_coords(x[i],y[i],0.,inc,PA)
    return x,y

def planet_position(model, i_planet, i_star, ):
    '''
     Returns planet position [arcsec] and PA [deg] in the map
    '''
    xy_planet = model.star_positions[:,0,0,i_planet]
    xy_star = model.star_positions[:,0,0,i_star]
    dxy = xy_planet - xy_star

    dist = np.hypot(dxy[0],dxy[1])
    PA = np.rad2deg(np.arctan2(dxy[1],-dxy[0])) + 360 - 90

    return [dist, PA]

def add_colorbar(mappable, shift=None, width=0.05, ax=None, trim_left=0, trim_right=0, side="right",**kwargs):
    # creates a color bar that does not shrink the main plot or panel
    # only works for horizontal bars so far

    if ax is None:
        ax = mappable.axes

    # Get current figure dimensions
    try:
        fig = ax.figure
        p = np.zeros([1,4])
        p[0,:] = ax.get_position().get_points().flatten()
    except:
        fig = ax[0].figure
        p = np.zeros([ax.size,4])
        for k, a in enumerate(ax):
            p[k,:] = a.get_position().get_points().flatten()
    xmin = np.amin(p[:,0]) ; xmax = np.amax(p[:,2]) ; dx = xmax - xmin
    ymin = np.amin(p[:,1]) ; ymax = np.amax(p[:,3]) ; dy = ymax - ymin

    if side=="top":
        if shift is None:
            shift = 0.2
        cax = fig.add_axes([xmin + trim_left, ymax + shift * dy, dx - trim_left - trim_right, width * dy])
        cax.xaxis.set_ticks_position('top')
        return fig.colorbar(mappable, cax=cax, orientation="horizontal",**kwargs)
    elif side=="right":
        if shift is None:
            shift = 0.05
        cax = fig.add_axes([xmax + shift*dx, ymin, width * dx, dy])
        cax.xaxis.set_ticks_position('top')
        return fig.colorbar(mappable, cax=cax, orientation="vertical",**kwargs)

def get_planet_r_az(disk_PA, inc, planet_r, planet_PA):
    # For a given disk PA and inc, and planet PA, and projected separation
    # gives the az to be passed to the planet_az option and the deprojected separation
    # input :
    # disk_PA, inc and  planet_PA in degrees
    # planet_r radius in arcsec or au
    # output :
    #  r in same unit as planet_r
    # planet_az in degrees (is value to be passed to mcfost)

    # test.get_planet_rPA does the opposite from a mcfost image

    dPA = planet_PA - disk_PA
    az = np.arctan(np.tan(np.deg2rad(dPA)) / np.cos(np.deg2rad(inc)))
    az = - np.rad2deg(az) # conversion to deg and correct convention for mcfost


    y_p = planet_r * np.sin(np.deg2rad(dPA))
    x_p = planet_r * np.cos(np.deg2rad(dPA))

    x = x_p
    y = y_p / np.cos(np.deg2rad(inc))

    r = np.hypot(x,y)

    # test :
    #mcfost.get_planet_r_az(62.5,50.2, 0.60521173 * 157.2, 11.619613647460938)
    # should give : (130.00000395126042, 62.500002877567475)

    return r, az



import glob
import numpy as np

def _word_to_bool(word):
    """convert a string to boolean according the first 2 characters."""
    _accepted_bool_prefixes = ("T", ".T")
    return word.upper().startswith(_accepted_bool_prefixes)

class Photons:
    pass


class Wavelengths:
    pass


class Physics:
    pass


class Dust:
    def __init__(self):
        self.component = []


class DustComponent:
    pass


class Grid:
    pass


class Map:
    pass


class Zone:
    def __init__(self):
        self.dust = []


class Mol:
    def __init__(self):
        self.molecule = []

class Atomic:
    def __init__(self):
        self.atom = []

class Molecule:
    pass

class Atom:
    pass

class Star:
    pass


class Simu:
    def __init__(self):
        self.version = float()


class Params:

    _minimum_version = 3.0

    def __init__(self, filename=None, **kwargs):

        # Instantiate empty classes and lists to store properties
        self.simu = Simu()
        self.phot = Photons()
        self.wavelengths = Wavelengths()
        self.map = Map()
        self.grid = Grid()
        self.zones = []
        self.mol = Mol()
        self.atomic = Atomic()
        self.stars = []

        self.filename = filename
        self._read(**kwargs)

    def _read(self):

        with open(self.filename, mode="rt") as file:
            f = []
            # Reading file and removing comments
            for line in file:
                # Skipping comments and empty lines
                if (not line.startswith("#")) and (len(line.strip()) > 0):
                    f += [line]
            f = iter(f)

        # -- Version of the parameter file --
        line = next(f).split()
        self.simu.version = float(line[0])
        if self.simu.version < self._minimum_version - 1e-3:
            print("Parameter file version is ", self.simu.version)
            raise Exception(
                'Parameter file version must be at least {ver:.2f}'.format(
                    ver=self._minimum_version
                )
            )

        # -- Number of photon packages --
        # to support float notations (e.g. "1.28e8" or "64000.0"),
        # we read as float but convert to int
        line = next(f).split()
        self.phot.nphot_T = int(float(line[0]))
        line = next(f).split()
        self.phot.nphot_SED = int(float(line[0]))
        line = next(f).split()
        self.phot.nphot_image = int(float(line[0]))

        # -- Wavelengths --
        line = next(f).split()
        self.wavelengths.n_wl = int(line[0])
        self.wavelengths.wl_min = float(line[1])
        self.wavelengths.wl_max = float(line[2])

        line = next(f).split()
        self.simu.compute_T = _word_to_bool(line[0])
        self.simu.compute_SED = _word_to_bool(line[1])
        self.simu.use_default_wl = _word_to_bool(line[2])

        line = next(f).split()
        self.wavelengths.file = line[0]

        line = next(f).split()
        self.simu.separate_contrib = _word_to_bool(line[0])
        self.simu.separate_pola = _word_to_bool(line[1])

        # -- Grid --
        line = next(f).split()
        self.grid.type = int(line[0])
        line = next(f).split()
        self.grid.n_rad = int(line[0])
        self.grid.nz = int(line[1])
        self.grid.n_az = int(line[2])
        self.grid.n_rad_in = int(line[3])

        # -- Maps --
        line = next(f).split()
        self.map.nx = int(line[0])
        self.map.ny = int(line[1])
        self.map.size = float(line[2])

        line = next(f).split()
        self.map.RT_imin = float(line[0])
        self.map.RT_imax = float(line[1])
        self.map.RT_ntheta = int(line[2])
        self.map.lRT_centered = _word_to_bool(line[3])

        line = next(f).split()
        self.map.RT_az_min = float(line[0])
        self.map.RT_az_max = float(line[1])
        self.map.RT_n_az = int(line[2])

        line = next(f).split()
        self.map.distance = float(line[0])

        line = next(f).split()
        self.map.PA = float(line[0])

        # -- Scattering method --
        if (self.simu.version < 4.1):
            line = next(f).split()
            self.simu.scattering_method = int(line[0])
        else:
            self.simu.scattering_method = 0

        line = next(f).split()
        self.simu.phase_function_method = int(line[0])

        # -- Symetries --
        line = next(f).split()
        self.simu.image_symmetry = _word_to_bool(line[0])
        line = next(f).split()
        self.simu.central_symmetry = _word_to_bool(line[0])
        line = next(f).split()
        self.simu.axial_symmetry = _word_to_bool(line[0])

        # -- Disk physics --
        line = next(f).split()
        self.simu.dust_settling_type = int(line[0])
        self.simu.dust_settling_exp = float(line[1])
        self.simu.a_settling = float(line[2])

        line = next(f).split()
        self.simu.radial_migration = _word_to_bool(line[0])

        line = next(f).split()
        self.simu.dust_sublimation = _word_to_bool(line[0])

        line = next(f).split()
        self.simu.hydrostatic_eq = _word_to_bool(line[0])

        line = next(f).split()
        self.simu.viscous_heating = _word_to_bool(line[0])
        self.simu.viscosity = float(line[1])

        # -- Number of zones --
        line = next(f).split()
        n_zones = int(line[0])
        self.simu.n_zones = n_zones

        # -- Density structure --
        for k in range(n_zones):
            self.zones.append(Zone())
            line = next(f).split()
            self.zones[k].geometry = int(line[0])
            line = next(f).split()
            self.zones[k].dust_mass = float(line[0])
            self.zones[k].gas_to_dust_ratio = float(line[1])

            line = next(f).split()
            self.zones[k].h0 = float(line[0])
            self.zones[k].Rref = float(line[1])
            self.zones[k].vertical_exp = float(line[2])

            line = next(f).split()
            self.zones[k].Rin = float(line[0])
            self.zones[k].edge = float(line[1])
            self.zones[k].Rout = float(line[2])
            self.zones[k].Rc = float(line[3])

            line = next(f).split()
            self.zones[k].flaring_exp = float(line[0])

            line = next(f).split()
            self.zones[k].surface_density_exp = float(line[0])
            self.zones[k].m_gamma_exp = float(line[1])

        # -- Grain properties --
        for k in range(n_zones):
            line = next(f).split()
            n_species = int(line[0])
            self.zones[k].n_species = n_species

            for j in range(n_species):
                self.zones[k].dust.append(Dust())

                line = next(f).split()
                self.zones[k].dust[j].type = line[0]
                n_components = int(line[1])
                self.zones[k].dust[j].n_components = n_components
                self.zones[k].dust[j].mixing_rule = int(line[2])
                self.zones[k].dust[j].porosity = float(line[3])
                self.zones[k].dust[j].mass_fraction = float(line[4])
                self.zones[k].dust[j].DHS_Vmax = float(line[5])

                for l in range(n_components):
                    self.zones[k].dust[j].component.append(DustComponent())

                    line = next(f).split()
                    self.zones[k].dust[j].component[l].file = line[0]
                    self.zones[k].dust[j].component[l].volume_fraction = float(line[1])

                line = next(f).split()
                self.zones[k].dust[j].heating_method = int(line[0])

                line = next(f).split()
                self.zones[k].dust[j].amin = float(line[0])
                self.zones[k].dust[j].amax = float(line[1])
                self.zones[k].dust[j].aexp = float(line[2])
                self.zones[k].dust[j].n_grains = int(line[3])


        # -- Molecular settings --
        line = next(f).split()
        self.mol.compute_pop = _word_to_bool(line[0])
        self.mol.compute_pop_accurate = _word_to_bool(line[1])
        self.mol.LTE = _word_to_bool(line[2])
        if (self.simu.version < 4.1):
            self.mol.profile_width = float(line[3])
        else:
            self.mol.profile_width = 15.0

        line = next(f).split()
        self.mol.v_turb = float(line[0])
        if (self.simu.version > 4.0):
            self.mol.v_turb_unit = line[1]
        else:
            self.mol.v_turb_unit = "km/s"

        line = next(f).split()
        n_mol = int(line[0])
        self.mol.n_mol = n_mol

        for k in range(n_mol):
            self.mol.molecule.append(Molecule())

            line = next(f).split()
            self.mol.molecule[k].file = line[0]
            self.mol.molecule[k].level_max = int(line[1])

            if (self.simu.version < 4.1):
                line = next(f).split()
                self.mol.molecule[k].v_max = float(line[0])
                self.mol.molecule[k].v_min = - self.mol.molecule[k].v_max
                self.mol.molecule[k].nv = 2*int(line[1])+1

            line = next(f).split()
            self.mol.molecule[k].cst_abundance = _word_to_bool(line[0])
            self.mol.molecule[k].abundance = line[1]
            self.mol.molecule[k].abundance_file = line[2]

            line = next(f).split()
            self.mol.molecule[k].ray_tracing = _word_to_bool(line[0])
            nTrans = int(line[1])
            self.mol.molecule[k].n_trans = nTrans

            line = next(f).split()
            self.mol.molecule[k].transitions = list(
                map(int, line[0:nTrans])
            )  # convert list of str to int

            if (self.simu.version > 4.0):
                line = next(f).split()
                self.mol.molecule[k].v_min = float(line[0])
                self.mol.molecule[k].v_max = float(line[1])
                self.mol.molecule[k].nv = int(line[2])


        if (self.simu.version > 3.0):
            # -- Atom settings --
            line = next(f).split()
            n_atoms = int(line[0])
            self.atomic.n_atoms = n_atoms

            for k in range(n_atoms):
                self.atomic.atom.append(Atom())

                line = next(f).split()
                self.atomic.atom[k].file = line[0]

                line = next(f).split()
                self.atomic.atom[k].nLTE = _word_to_bool(line[0])

                line = next(f).split()
                self.atomic.atom[k].initial_solution = int(line[0])

                line = next(f).split()
                self.atomic.atom[k].v_max = float(line[0])
                self.atomic.atom[k].nv = int(line[1])

                line = next(f).split()
                self.atomic.atom[k].images = _word_to_bool(line[0])
                n_trans = int(line[1])
                self.atomic.atom[k].n_trans = n_trans

                self.atomic.atom[k].lower = np.zeros(n_trans, dtype=int)
                self.atomic.atom[k].upper = np.zeros(n_trans, dtype=int)
                for l in range(n_trans):
                    line = next(f).split()
                    self.atomic.atom[k].lower[l] = int(line[0])
                    self.atomic.atom[k].upper[l] = int(line[1])
        else:
            self.atomic.n_atoms = 0

        # -- Star properties --
        line = next(f).split()
        n_stars = int(line[0])
        self.simu.n_stars = n_stars

        for k in range(n_stars):
            self.stars.append(Star())

            line = next(f).split()
            self.stars[k].Teff = float(line[0])
            self.stars[k].R = float(line[1])
            self.stars[k].M = float(line[2])
            self.stars[k].x = float(line[3])
            self.stars[k].y = float(line[4])
            self.stars[k].z = float(line[5])
            self.stars[k].is_bb = _word_to_bool(line[6])

            line = next(f).split()
            self.stars[k].file = line[0]

            line = next(f).split()
            self.stars[k].fUV = float(line[0])
            self.stars[k].slope_UV = float(line[1])

        # -- Command line options --
        for line in f:
            if (len(line) > 0):
                line = line.split()
                if (len(line) > 0): # we test again in case there were only spaces
                    if (line[0] == "Executed"):
                        self.options = " ".join(line[6:])

                        if (line[0] == "sha"):
                            self.mcfost_sha = line[2]



    def __str__(self):
        """ Return a formatted parameter file. Currently returns v3.0 format
        """

        # -- Photon packets --
        txt = f"""4.1                       mcfost version\n
#-- Number of photon packages --
  {self.phot.nphot_T:<10.5g}              nbr_photons_eq_th  : T computation
  {self.phot.nphot_SED:<10.5g}              nbr_photons_lambda : SED computation
  {self.phot.nphot_image:<10.5g}              nbr_photons_image : images computation\n\n"""

        # -- Wavelengths --
        txt += f"""#-- Wavelength --
  {self.wavelengths.n_wl:<4d} {self.wavelengths.wl_min:<7.3f} {self.wavelengths.wl_max:<7.3f}      n_lambda, lambda_min, lambda_max [microns]
  {self.simu.compute_T} {self.simu.compute_SED} {self.simu.use_default_wl}         compute temperature?, compute sed?, use default wavelength grid ?
  {self.wavelengths.file}            wavelength file (if previous parameter is F)
  {self.simu.separate_contrib} {self.simu.separate_pola}              separation of different contributions?, stokes parameters?\n\n"""

        # -- Grid --
        txt += f"""#-- Grid geometry and size --
  {self.grid.type:>1d}                       1 = cylindrical, 2 = spherical
  {self.grid.n_rad} {self.grid.nz} {self.grid.n_az} {self.grid.n_rad_in}             n_rad (log distribution), nz (or n_theta), n_az, n_rad_in\n\n"""

        # -- Maps --
        txt += f"""#-- Maps --
  {self.map.nx} {self.map.ny} {self.map.size:5.1f}           grid (nx,ny), size [au]
  {self.map.RT_imin:<4.1f}  {self.map.RT_imax:<4.1f}  {self.map.RT_ntheta:>2d} {self.map.lRT_centered}    RT: imin, imax, n_incl, centered ?
  {self.map.RT_az_min:<4.1f}  {self.map.RT_az_max:<4.1f}  {self.map.RT_n_az:>2d}          RT: az_min, az_max, n_az
  {self.map.distance:<6.2f}                  distance (pc)
  {self.map.PA:<6.2f}                  disk PA\n\n"""

        # -- Scattering method --
        txt += f"""#-- Scattering method --
  {self.simu.phase_function_method}                       1=Mie, 2=hg (2 implies the loss of polarizarion)\n\n"""

        # -- Symetries --
        txt += f"""#-- Symmetries --
  {self.simu.image_symmetry}                    image symmetry
  {self.simu.central_symmetry}                    central symmetry
  {self.simu.axial_symmetry}                    axial symmetry (important only if N_phi > 1)\n\n"""

        # -- Disk physics --
        txt += f"""#Disk physics
  {self.simu.dust_settling_type}  {self.simu.dust_settling_exp:<6.2f}  {self.simu.a_settling:<6.2f}       dust_settling (0=no settling, 1=parametric, 2=Dubrulle, 3=Fromang), exp_strat, a_strat (for parametric settling)
  {self.simu.radial_migration}                   dust radial migration
  {self.simu.dust_sublimation}                   sublimate dust
  {self.simu.hydrostatic_eq}                   hydrostatic equilibrium
  {self.simu.viscous_heating}  {self.simu.viscosity:4.1g}            viscous heating, alpha_viscosity\n\n"""

        # -- Number of zones --
        txt += f"""#-- Number of zones --   1 zone = 1 density structure + corresponding grain properties
  {self.simu.n_zones}\n\n"""

        # -- Density structure --
        txt += f"#-- Density structure --\n"
        for k in range(self.simu.n_zones):
            txt += f"""  {self.zones[k].geometry}                        zone type : 1 = disk, 2 = tapered-edge disk, 3 = envelope, 4 = debris disk, 5 = wall
  {self.zones[k].dust_mass} {self.zones[k].gas_to_dust_ratio}         dust mass,  gas-to-dust mass ratio
  {self.zones[k].h0}  {self.zones[k].Rref} {self.zones[k].vertical_exp}     scale height, reference radius (AU), unused for envelope, vertical profile exponent (only for debris disk)
  {self.zones[k].Rin}  {self.zones[k].edge} {self.zones[k].Rout} {self.zones[k].Rc}  Rin, edge, Rout, Rc (AU) Rc is only used for tappered-edge & debris disks (Rout set to 8*Rc if Rout==0)
  {self.zones[k].flaring_exp}                 flaring exponent, unused for envelope
  {self.zones[k].surface_density_exp} {self.zones[k].m_gamma_exp}                 surface density exponent (or -gamma for tappered-edge disk or volume density for envelope), usually < 0, -gamma_exp (or alpha_in & alpha_out for debris disk)\n\n"""
        txt += f"\n"

        # -- Grain properties --
        txt += f"#-- Grain properties --\n"
        for k in range(self.simu.n_zones):
            txt += (
                f"  {self.zones[k].n_species}                      Number of species\n"
            )
            for j in range(self.zones[k].n_species):
                txt += f"  Mie {self.zones[k].dust[j].n_components} {self.zones[k].dust[j].mixing_rule} {self.zones[k].dust[j].porosity:<5.2f} {self.zones[k].dust[j].mass_fraction:<5.2f} {self.zones[k].dust[j].DHS_Vmax}    Grain type (Mie or DHS), N_components, mixing rule (1 = EMT or 2 = coating),  porosity, mass fraction, Vmax (for DHS)\n"
                for l in range(self.zones[k].dust[j].n_components):
                    txt += f"  {self.zones[k].dust[j].component[l].file}  {self.zones[k].dust[j].component[l].volume_fraction}     Optical indices file, volume fraction\n"
                txt += f"""  {self.zones[k].dust[j].heating_method}                          Heating method : 1 = RE + LTE, 2 = RE + NLTE, 3 = NRE
  {self.zones[k].dust[j].amin} {self.zones[k].dust[j].amax}  {self.zones[k].dust[j].aexp} {self.zones[k].dust[j].n_grains}       amin, amax, aexp, nbr_grains\n\n"""

        # -- Molecular settings --
        txt += f"""#-- Molecular RT settings --
  {self.mol.compute_pop}  {self.mol.compute_pop_accurate}  {self.mol.LTE} {self.mol.profile_width}      lpop, laccurate_pop, LTE, profile width
  {self.mol.v_turb} {self.mol.v_turb_unit}                       v_turb, unit
  {self.mol.n_mol}                          nmol\n"""
        for k in range(self.mol.n_mol):
            txt += f"""  {self.mol.molecule[k].file} {self.mol.molecule[k].level_max}          molecular data filename, level_max
  {self.mol.molecule[k].cst_abundance} {self.mol.molecule[k].abundance} {self.mol.molecule[k].abundance_file}   cst molecule abundance ?, abundance, abundance file
  {self.mol.molecule[k].ray_tracing}  {self.mol.molecule[k].n_trans}                   ray tracing ?,  number of lines in ray-tracing\n """
            for j in range(self.mol.molecule[k].n_trans):
                txt += f" {self.mol.molecule[k].transitions[j]}"
            txt += f" transition numbers\n"
        txt += f"{self.mol.molecule[k].v_min} {self.mol.molecule[k].v_max} {self.mol.molecule[k].nv}                 vmin, vmax (km.s-1), n_speed\n"

        txt += f"\n"

         # -- Atomic settings --
        txt += f"""#-- Atomic RT settings --
  {self.atomic.n_atoms}   number of atoms\n"""
        for k in range(self.atomic.n_atoms):
            txt += f"""  {self.atomic.atom[k].file}          atomic data filename, level_max
  {self.atomic.atom[k].nLTE}          non-LTE ?
  {self.atomic.atom[k].initial_solution}          initial solution : 0 = LTE, 1 = from file
  {self.atomic.atom[k].v_max} {self.atomic.atom[k].nv}                 vmax (km.s-1), n_speed
  {self.atomic.atom[k].images} {self.atomic.atom[k].n_trans}                   images (T) or total flux (F)?,  number of lines in ray-tracing\n """
            for j in range(self.atomic.atom[k].n_trans):
                txt += f" {self.atomic.atom[k].lower[j]} {self.atomic.atom[k].upper[j]}    lower and upper levels\n"
        txt += f"\n"

        # -- Star properties --
        txt += f"""#-- Star properties --
  {self.simu.n_stars}  Number of stars\n"""
        for k in range(self.simu.n_stars):
            txt += f"""  {self.stars[k].Teff} {self.stars[k].R} {self.stars[k].M} {self.stars[k].x} {self.stars[k].y} {self.stars[k].x} {self.stars[k].is_bb}  Temp, radius (solar radius),M (solar mass),x,y,z (AU), is a blackbody?
  {self.stars[k].file}
  {self.stars[k].fUV} {self.stars[k].slope_UV}     fUV, slope_UV\n"""

        return txt

    def writeto(self, outname):
        """ Write an MCFOST parameter file to disk.  """
        with open(outname, mode="wt") as file:
            file.write(str(self))

    def calc_inclinations(self):
        # Calculate the inclinations for the ray-traced SEDs and images
        if self.map.RT_ntheta == 1:
            return self.map.RT_imin
        else:
            cos_min, cos_max = np.cos(np.deg2rad([self.map.RT_imin, self.map.RT_imax]))
            if self.map.lRT_centered:
                return (
                    np.rad2deg(np.arccos(
                        cos_min
                        + (np.arange(self.map.RT_ntheta) + 0.5)
                        / self.map.RT_ntheta
                        * (cos_max - cos_min)
                    ))
                )
            else:
                return (
                    np.rad2deg(np.arccos(
                        cos_min
                        + (np.arange(self.map.RT_ntheta))
                        / (self.map.RT_ntheta - 1)
                        * (cos_max - cos_min)
                    ))
                )


def find_parameter_file(directory="./"):

    list = glob.glob(directory + "/*.par*")
    if len(list) == 1:
        return list[0]
    elif len(list) > 1:
        raise ValueError("Multiple parameter files found in " + directory)
    else:
        raise ValueError("No parameter files found in " + directory)



import numpy as np
import astropy.io.fits as fits
import astropy.units as u
import sys
import subprocess
from .parameters import Params
from .disc_structure import Disc


def fargo2mcfost(
    data_dir,
    dump_number,
    nz=5,
    mcfost_ref_file="/Users/cpinte/mcfost/src/ref3.0_3D.para",
    mcfost_filename="mcfost_FARGO.para",
    fitsname=None,
):

    # --- Need to read fargo .par file here, hard-coded for now
    # --- TODO : read Frame == C ou F  to detect is planet is rotating or not
    # --- if C : add keplrian velocity of planet
    # TODO : we need to read FACTORUNITMASS & FACTIORUNITLENGHT
    # -- MM
    Runit = 10.0
    Munit = 1.0  # star always has a mass == 1
    Frame = "C"
    Rmin = 4
    Rmax = 25

    # -- TODO : easiest if probably to read Rmin, Rmax, Nrad, Nsec here too
    # -- then we can skip dims.dat

    # --- Load dims.dat
    # This is an output file which contains for example the number of grid points
    try:
        dims = np.loadtxt(data_dir + '/dims.dat')
    except:
        sys.exit('Could not load dims.dat')
    N_t = int(dims[5])  # Number of outputs
    N_R = int(dims[6])  # Number of radial gridpoint
    N_theta = int(dims[7])  # Number of angular grid points

    # --- Load planet0.dat
    try:
        planet0 = np.loadtxt(data_dir + '/planet0.dat')
    except:
        sys.exit('Could not load planet0.dat!')
    dummy_i = -1

    # -- Todo : I have have no idea what those lines are doing
    # -- the test planet0[i, 0] == dummy_i is always false
    for i in range(planet0.shape[0]):
        if planet0[i, 0] == dummy_i:
            planet0[i - 1 : planet0.shape[0] - 1, :] = planet0[i:, :]
        dummy_i = planet0[i, 0]

    # Cartesian coordinates of the planet
    planet_x = planet0[dump_number, 1] * u.AU
    planet_y = planet0[dump_number, 2] * u.AU
    # Transform to polar coordinates
    planet_R = np.sqrt(planet_x ** 2 + planet_y ** 2)
    planet_theta = np.arctan2(planet_y, planet_x)

    # --- physical ctes
    grav = 39.4862194  # (au^3) / (solar mass * (yr^2))
    Timeunit = np.sqrt(Runit ** 3.0 / (grav * Munit))  # yr
    Lumunit = (
        Munit * Runit * Runit / (Timeunit * Timeunit * Timeunit)
    )  #  Msun*au^2*yr^-3
    Lsun = 2.7e-4  # Msun au^2/yr^3
    Lestrella = 1.0 * (Lsun / Lumunit)  #
    Rgas = (
        3.67e-4
    )  # //au^2/(yr^2 K)   R =0.000183508935; // R = 4124 [J/kg/K] = 0.000183508935 (au^2) / ((year^2) * kelvin)
    Tempunit = 2.35 * (grav * Munit / (Rgas * Runit))  # Kelvin
    densINcgs = 8888035.76  # 1 Msun/au^2 = 8888035.76 grams / (cm^2)
    DensUnit = densINcgs * Munit / Runit ** 2.0  # grams / (cm^2)
    QplusUnit = 1.6e13  # erg/s/cm^2
    vel_unit = Runit / Timeunit * 474057.172  # 1 au/yr = 474 057.172 cm/s

    # --- Does the planet rotate
    if Frame == "C":
        vtheta_planet = np.sqrt(grav * Munit / planet_R) * vel_unit
    else:
        vtheta_planet = 0.0

    # --- Reading data
    rho = np.fromfile(
        data_dir + "/gasdens{0:d}.dat".format(dump_number), dtype='float64'
    )
    temp = np.fromfile(
        data_dir + "/gasTemperature{0:d}.dat".format(dump_number), dtype='float64'
    )
    vrad = np.fromfile(
        data_dir + "/gasvrad{0:d}.dat".format(dump_number), dtype='float64'
    )
    vtheta = np.fromfile(
        data_dir + "/gasvtheta{0:d}.dat".format(dump_number), dtype='float64'
    )

    # --- Correct units + add velocity offset if needed
    rho_cgs = DensUnit * rho  # gr/cm^2
    temp_cgs = Tempunit * temp  # Kelvin
    vrad_cgs = vel_unit * vrad
    vtheta_cgs = vel_unit * vtheta + vtheta_planet.value

    # --- reshape
    Rho = DensUnit * rho.reshape(N_R, N_theta)
    Temp = Tempunit * temp.reshape(N_R, N_theta)

    # --- mcfost : updating values from FARGO
    # --- Setting up a parameter file and creating the corresponding grid
    P = Params(mcfost_ref_file)
    P.grid.n_rad = N_R
    P.grid.n_rad_in = 1
    P.grid.nz = nz
    P.grid.n_az = N_theta

    P.zones[0].Rin = Rmin
    P.zones[0].edge = 0.0
    P.zones[0].Rout = Rmax
    P.zones[0].Rc = 0.0

    # --- TODO : compute spectral type for given stellar mass
    # -- CP

    # -- Turn off symmetries
    P.simu.image_symmetry = False
    P.simu.central_symmetry = False
    P.simu.axial_symmetry = False

    # -- TODO : compute gas mass properly and put it in mcfost_FARGO.para
    # -- MM

    # -- Write new parameter file
    P.writeto(mcfost_filename)

    # --- Running mcfost to create the grid
    subprocess.call(["rm", "-rf", "data_disk", "data_disk_old"])
    result = subprocess.call(["mcfost", mcfost_filename, "-disk_struct"])

    # --- Reading mcfost grid
    mcfost_disc = Disc("./")

    # --- print the mcfost radial grid to check that it is the same as FARGO's
    print("MCFOST radii=")
    print(mcfost_disc.r()[0, 0, :])

    # -- same dimension order as FARGO
    mcfost_z = mcfost_disc.z()

    # --taking only half the grid (+z)
    mcfost_z = mcfost_z[:, nz + 1 :, :]
    mcfost_z = mcfost_z.transpose()  # dims are now, r, z, theta

    # -- TODO: defining H [au]
    # --- MM
    h = 0.05 * mcfost_disc.r()[:, 0, :].transpose()

    # -- computing the 3D density structure for mcfost
    rho_mcfost = Rho[:, np.newaxis, :] * np.exp(-0.5 * (mcfost_z / h[:, np.newaxis, :]))

    # --- Write a fits file for mcfost
    if fitsname is not None:
        print("Writing ", fitsname, " to disk")
        fits.writeto(fitsname, rho_mcfost.transpose(), overwrite=True)
        # --- TODO : add velocity : CP
    else:
        print("Fits file for mcfost was not created")

    return rho_mcfost

# Run with :
# import pymcfost as mcfost
# rho = mcfost.fargo2mcfost("/Users/cpinte/Downloads/4mcfost/",2,fitsname="test.fits")
# plot density with plt.imshow(rho[:,0,:])
# run mcfost with : mcfost_FARGO.para -df test.fits  # computes temperature & SED
#  mcfost_FARGO.para -df test.fits -img 1300 # creates an image at 1.3mm



import astropy.io.fits as fits
import matplotlib.pyplot as plt
import matplotlib.tri as tri
import numpy as np
import os

from .parameters import Params, find_parameter_file
from .run import run

class Disc:
    def __init__(self, dir=None, **kwargs):
        # Correct path if needed
        dir = os.path.normpath(os.path.expanduser(dir))
        if dir[-9:] != "data_disk":
            dir = os.path.join(dir, "data_disk")
        self.dir = dir

        # Search for parameter file
        para_file = find_parameter_file(dir)

        # Read parameter file
        self.P = Params(para_file)

        # Read model results
        self._read(**kwargs)

    def _read(self):
        # Read grid file
        try:
            hdu = fits.open(self.dir + "/grid.fits.gz")
            self.grid = hdu[0].data
            hdu.close()
        except OSError:
            print('cannot open grid.fits.gz')

        # Read gas density file
        try:
            hdu = fits.open(self.dir + "/gas_density.fits.gz")
            self.gas_density = hdu[0].data
            hdu.close()
        except OSError:
            print('cannot open gas_density.fits.gz')

        # Read volume file
        try:
            hdu = fits.open(self.dir + "/volume.fits.gz")
            self.volume = hdu[0].data
            hdu.close()
        except OSError:
            print('cannot open volume.fits.gz')

    def r(self):
        if self.grid.ndim > 2:
            return self.grid[0, :, :, :]
        else:
            return np.sqrt(self.grid[0, :] ** 2 + self.grid[1, :] ** 2)

    def z(self):
        if self.grid.ndim > 2:
            return self.grid[1, :, :, :]
        else:
            return self.grid[2, :]

    def add_spiral(
        self, a=30, sigma=10, f=1, theta0=0, rmin=None, rmax=None, n_az=None
    ):
        """ Add a geometrucal spiral on a 2D (or 3D) mcfost density grid
        and return a 3D array which can be directly written as a fits
        file for mcfost to read

        geometrical spiral r = a (theta - theta0)
        surface density is mutiply by f at the crest of the spiral
        the spiral has a Gaussin profil in (x,y) with sigma given in au
        """

        if self.grid.ndim <= 2:
            ValueError("Can only add a spiral on a cylindrical or spherical grid")

        if n_az is None:
            n_az = self.grid.shape[1]
        phi = np.linspace(0, 2 * np.pi, n_az, endpoint=False)

        r = self.grid[0, 0, 0, :]

        if rmin is None:
            rmin = r.min()
        if rmax is None:
            rmax = r.max()

        x = r[np.newaxis, :] * np.cos(phi[:, np.newaxis])
        y = r[np.newaxis, :] * np.sin(phi[:, np.newaxis])

        # Just to test
        # x = np.linspace(-100,100,500)
        # x, y = np.meshgrid(x,x)
        # r = np.sqrt(x**2 + y**2) # we recalcule in preparation for other types of grid

        # rc=50, hc=0.15, alpha=1.5, beta=0.25
        # theta_c = 0.
        # theta = theta_c + np.sign(r - rc)/hc * \
        #        ((r/rc)**(1+beta) * (1/(1+beta) - 1/(1-alpha + beta) * (r/rc)**(-alpha)) \
        #         - 1/(1+beta) - 1/(1-alpha + beta))

        r_spiral = np.geomspace(rmin, rmax, num=5000)
        theta = r_spiral / a + theta0

        x_spiral = r_spiral * np.cos(theta)
        y_spiral = r_spiral * np.sin(theta)

        correct = np.ones(x.shape)

        # This is really badly implemented, but fast enough that we don't care
        sigma2 = sigma ** 2
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                d2 = np.min((x_spiral - x[i, j]) ** 2 + (y_spiral - y[i, j]) ** 2)
                correct[i, j] += f * np.exp(-0.5 * d2 / sigma2)

        triang = tri.Triangulation(x.flatten(), y.flatten())
        plt.tripcolor(triang, correct.flatten(), shading='flat')

        return self.gas_density[np.newaxis, :, :] * correct[:, np.newaxis, :]



def check_grid(model):
    """
    We check if the disc structure already exists
    if not, we check if it exists
    if not, we try to compute it
    """

    try:
        grid = model.disc.grid
    except:
        try:
            print("Trying to read grid structure ...")
            model.disc = Disc(model.basedir)
            grid = model.disc.grid
        except:
            print("No grid structure, trying to create it ...")
            run(model.P.filename, options=model.P.options+" -disk_struct")
            try:
                print("Trying to read grid structure again ...")
                model.disc = Disc(model.basedir)
                grid = model.disc.grid
            except AttributeError:
                print("Cannot read grid in " + model.basedir)

    return grid


def _plot_cutz(model, y, r=None, dr=None, log=None, **kwargs):

    grid = check_grid(model)

    if grid.ndim > 2:
        r_mcfost = grid[0, 0, 0, :]
        i = np.argmin(np.abs(r_mcfost - r))
        print("selected_radius =", r_mcfost[i])
        z_mcfost = grid[1, 0, :, i]

        y = y[:,i]

        if log:
            plt.loglog(z_mcfost, y, **kwargs)
        else:
            plt.plot(z_mcfost, y, **kwargs)

    else:
        r_mcfost = np.sqrt(grid[0, :] ** 2 + grid[1, :] ** 2)
        ou = r_mcfost > 1e-6  # Removing star
        y = y[ou]
        r_mcfost = r_mcfost[ou]
        z_mcfost = grid[2, ou]

        # Selecting data points
        ou = (r_mcfost > r - dr) & (r_mcfost < r + dr)

        z_mcfost = z_mcfost[ou]
        y = y[ou]

        #plt.plot(z_mcfost, T, "o", **kwargs)
        fig = plt.gcf()
        ax = fig.add_subplot(1, 1, 1, projection='scatter_density')
        density = ax.scatter_density(z_mcfost,y, **kwargs)

    plt.xlabel("z [au]")



import numpy as np
import os
import subprocess
import scipy.constants as sc
from scipy import interpolate
from .image import Image
from .utils import Wm2_to_Jy, FWHM_to_sigma
from astropy.io import fits
from astropy.convolution import Gaussian2DKernel, convolve_fft, convolve
from scipy.ndimage import convolve1d

def pseudo_CASA_simdata(model,i=0,iaz=0,iTrans=None,simu_name = "pseudo_casa",beam=None,bmaj=None,bmin=None,bpa=None,subtract_cont=False,Delta_v=None, rms=0):
    """
    Generate a fits file as if it was a CASA simdata output
     - convolve spatially with beam (bmin, bmaj in arsec, bpa in degrees)
     - convolve spectrally with Delta_v in km/s
     - todo : resample in velocity as required

    Basically generate a CASA fits file with perfect uv coverage and no noise
    """

    workdir = "CASA/"
    if not os.path.exists(workdir):
        os.mkdir(workdir)
    _CASA_clean(workdir)

    is_image = isinstance(model, Image)

    # --- Checking arguments
    if not is_image:
        if iTrans is None:
            raise Exception("Missing transition number iTrans : iTrans is a python index from 0 to nTrans-1")

        nTrans = model.freq.size
        if iTrans > nTrans - 1:
            raise Exception(f"ERROR: iTrans is not in the computed range : nTrans={nTrans}")

    if beam is not None:
        bmaj = beam
        bmin = beam
        bpa = 0

    if bmaj is None:
        raise Exception("Missing beam")

    # -- Flux setup : we want Jy/pixel first
    if is_image:
        if model.is_casa:
            image = model.image[:, :]
        else:
            # Convert to Jy
            image = Wm2_to_Jy(model.image[0, iaz, i, :, :], sc.c / model.wl)
    else:  # cube
        if model.is_casa:
            # -- continuum subtraction
            if subtract_cont:
                image = model.lines[:,:,:] - model.lines[0,:,:]
            else:
                image = model.lines[:, :, :]
        else:
            # -- continuum subtraction
            if subtract_cont:
                image = model.lines[iaz, i, iTrans,:,:,:] - model.cont[iaz, i, iTrans, np.newaxis, :, :]
            else:
                image = model.lines[iaz, i, iTrans,:,:,:]

            # Convert to Jy
            image = Wm2_to_Jy(image, model.freq[iTrans])

    #-- writing fits file
    hdr = fits.Header()
    hdr["EXTEND"] = True
    hdr["OBJECT"] = "mcfost"
    hdr["CTYPE1"] = "RA---TAN"
    hdr["CRVAL1"] = 0.0
    hdr["CRPIX1"] = int(model.nx / 2 + 1)
    hdr["CDELT1"] = -model.pixelscale / 3600.0
    hdr["CUNIT1"] = "deg"

    hdr["CTYPE2"] = "DEC--TAN"
    hdr["CRVAL2"] = 0.0
    hdr["CRPIX2"] = int(model.ny / 2 + 1)
    hdr["CDELT2"] = model.pixelscale / 3600.0
    hdr["CUNIT2"] = "deg"

    if is_image:
        # 3rd axis
        hdr["CTYPE3"] = "STOKES"
        hdr["CRVAL3"] = 1.0
        hdr["CDELT3"] = 1.0
        hdr["CRPIX3"] = 1

        # 4th axis
        hdr["CTYPE4"] = "FREQ"
        hdr["CRVAL4"] = model.freq # Hz
        hdr["CDELT4"] = 2e9  # 2GHz by default
        hdr["CRPIX4"] = 0
    else:
        hdr["CTYPE3"] = "VELO-LSR"
        hdr["CRVAL3"] = 0.0  # line center
        hdr["CRPIX3"] = model.nv//2 + 1
        hdr["CDELT3"] = model.dv * 1000
        hdr["CUNIT3"] = "m/s"

    hdr["RESTFREQ"] = model.freq[iTrans]  # Hz
    hdr["BUNIT"] = "JY/BEAM"
    hdr["BTYPE"] = "Intensity"
    hdr["BMAJ"] = bmaj/3600.
    hdr["BMIN"] = bmin/3600.
    hdr["BPA"] = bpa

    # Convolve spectrally
    if Delta_v is not None:
        image = model._spectral_convolve(image, Delta_v)

    print(f"Spatial convolution at {bmaj} x {bmin}")
    #-- Convolution by beam
    sigma_x = bmin / model.pixelscale * FWHM_to_sigma  # in pixels
    sigma_y = bmaj / model.pixelscale * FWHM_to_sigma  # in pixels
    beam = Gaussian2DKernel(sigma_x, sigma_y, bpa * np.pi / 180)

    if image.ndim == 2:
        image = convolve_fft(image, beam)
    else:
        for iv in range(image.shape[0]):
            image[iv,:,:] = convolve_fft(image[iv,:,:], beam)

    #-- Jy/pixel to Jy/beam
    beam_area = bmin * bmaj * np.pi / (4.0 * np.log(2.0))
    pix_area = model.pixelscale**2
    image *= beam_area/pix_area

    print(f"Peak flux is {np.max(image)} Jy/beam")

	#-- This is for testing purpose only so far: this needs to be updated and to come before
	#--  - compute the scale factor in 1 channel once,
	#--  - then add noise before spatial and spectral convolution so we do not convolve twice
    if rms > 0.0:
        noise = np.random.randn(image.size).reshape(image.shape)
        for iv in range(image.shape[0]):
            noise[iv,:,:] = convolve_fft(noise[iv,:,:], beam)
        if Delta_v is not None:
            noise =  model._spectral_convolve(noise, Delta_v)
        #print(np.std(noise), beam_area/pix_area)
        noise *= rms / np.std(noise)
        image += noise

    hdu = fits.PrimaryHDU(image, header=hdr)
    hdul = fits.HDUList(hdu)

    hdul.writeto(workdir + simu_name + ".fits", overwrite=True)

#	if rms > 0.0:
#		noise = np.random.randn(cube.size).reshape(cube.shape)
#		noise = np.array([convolve(c, beam) for c in noise])
#		noise = np.convolve(noise, spectral_response, axis=0)
#		noise *= rms / np.std(noise)
#		image += noise

def CASA_simdata(
    model,
    i=0,
    iaz=0,
    obstime=None,
    config=None,
    resol=None,
    sampling_time=None,
    pwv=0.0,
    decl="-22d59m59.8",
    phase_noise=False,
    name="simu",
    iTrans=None,
    rt=True,
    only_prepare=False,
    interferometer='alma',
    mosaic=False,
    mapsize=None,
    channels=None,
    width=None,
    correct_flux=1.0,
    simu_name=None,
    ms=None,
    n_iter=10000,
    hourangle="transit"):
    """
    Prepare a MCFOST model for the CASA alma simulator

    Generates 2 files:
      - a python CASA script
      - a fits file with the required dimensions and  keywords

    Then run the simulator and export a fits file with the results

    Tested to work with CASA 5.4.0-68 on MacOS : command line to call CASA is assumed to be "casa"
    """

    workdir = "CASA/"
    if not os.path.exists(workdir):
        os.mkdir(workdir)
    _CASA_clean(workdir)

    is_image = isinstance(model, Image)

    # --- Checking arguments
    if not is_image:
        if iTrans is None:
            raise Exception("Missing transition number iTrans")

        nTrans = model.freq.size
        if iTrans > nTrans - 1:
            raise Exception(f"ERROR: iTrans is not in the computed range : nTrans={nTrans}")

    if ms is None:
        # --- Setting a configuration and observing time for simalma
        simobs_custom = False

        if obstime is None:
            raise Exception("Missing obstime")

        if sampling_time is None:
            sampling_time = obstime / 1000

        if config is None:
            if resol is None:
                raise Exception("Missing config or resol")
            else:
                resol_name = f"{resol:2.2f}"
                resol_name_script = f"alma_={resol:6.6f}arcsec"
        else:
            if isinstance(config, int):
                config = f"alma.cycle6.{config}"
                resol_name = "_config=" + config
                resol_name_script = config
            else: # we do not change config
                resol_name_script = config

    else:
        # -- Setting up for simobs_custom
        simobs_custom = True

    # -- Thermal noise
    if pwv is None:
        print("pwv not specified --> No thermal noise")
        th_noise = "''"
        lth_noise = 0
    else:
        th_noise = "'tsys-atm'"
        spwv = f"{pwv:4.2f}"
        is_th_noise = True

    # -- Frequency setup
    if is_image:
        freq = sc.c / (model.wl * 1e-6) * 1e-9  # [Ghz]
        inwidth = 8  # 8 Ghz par defaut en continu
        print(f"Setting channel width to 8Ghz")
        inchan = 1  # 1 channel pour continu
    else:  # cube
        dv = model.dv * 1000.0  # [m/s]
        freq = model.freq[iTrans] * 1e-9  # [Ghz]

        if width is None:
            inwidth = dv / sc.c * model.freq[iTrans] * 1e-9  # [Ghz]
            print(f"Setting channel width to {dv:f} m/s")
        else:
            inwidth = width

        if channels is None:
            inchan = 2 * model.P.mol.molecule[imol].nv + 1
            channels = np.arange(inchan)
        else:
            if isinstance(channels, int):
                inchan = 1
            else:
                inchan = len(channels)

    # -- Flux setup
    if is_image:
        if model.is_casa:
            image = model.image[:, :]
        else:
            image = Wm2_to_Jy(model.image[0, iaz, i, :, :], sc.c / model.wl)  # Convert to Jy
            image = image[
                np.newaxis, np.newaxis, :, :
            ]  # Adding spectral & pola dimensions
    else:  # cube
        if model.is_casa:
            image = model.lines[channels, :, :]
        else:
            image = Wm2_to_Jy(model.lines[iaz, i, iTrans, channels, :, :], model.freq[iTrans])  # Convert to Jy
    if (image.ndim == 2):  # Adding extra spectral dimension if there is only 1 channel selected
        image = image[np.newaxis, :, :]

    # -- pixels
    incell = model.pixelscale

    # -- Filenames
    if simu_name is None:
        simu_name = "casa_simu"
    if is_image:
        simu_name = simu_name + f"_lambda={model.wl}_obstime={obstime}_decl=" + decl

    # ---------------------------------------------
    # -- fits file
    # ---------------------------------------------

    hdr = fits.Header()
    hdr["EXTEND"] = True
    hdr["CTYPE1"] = "RA---TAN"
    hdr["CRVAL1"] = 0.0
    hdr["CRPIX1"] = int(model.nx / 2 + 1)
    hdr["CDELT1"] = -model.pixelscale / 3600.0

    hdr["CTYPE2"] = "DEC--TAN"
    hdr["CRVAL2"] = 0.0
    hdr["CRPIX2"] = int(model.ny / 2 + 1)
    hdr["CDELT2"] = model.pixelscale / 3600.0

    if is_image:
        # 3rd axis
        hdr["CTYPE3"] = "STOKES"
        hdr["CRVAL3"] = 1.0
        hdr["CDELT3"] = 1.0
        hdr["CRPIX3"] = 1

        # 4th axis
        hdr["CTYPE4"] = "FREQ"
        hdr["CRVAL4"] = freq * 1e9  # Hz
        hdr["CDELT4"] = 2e9  # 2GHz by default
        hdr["CRPIX4"] = 0
    else:
        # WARNING this is incorrect : the simulator will still work but the velocity axis in the output file will be off
        hdr["CTYPE3"] = "VELO-LSR"
        hdr["CRVAL3"] = 0.0  # line center
        hdr["CRPIX3"] = inchan
        hdr["CDELT3"] = dv

    hdr["RESTFREQ"] = freq * 1e9  # Hz
    hdr["BUNIT"] = "JY/PIXEL"
    hdr["BTYPE"] = "Intensity"

    hdu = fits.PrimaryHDU(image, header=hdr)
    hdul = fits.HDUList(hdu)
    hdul.writeto(workdir + simu_name + ".raw.fits", overwrite=True)

    # ---------------------------------------------
    # -- CASA script
    # ---------------------------------------------
    # spatial setup
    txt = f"""project = 'DISK'
skymodel = '{simu_name}.raw.fits'
dryrun = False
modifymodel = True
inbright = 'unchanged'
indirection = 'J2000 18h00m00.02 {decl}' # mosaic center, or list of pointings
hourangle = '{hourangle}'
incell = '{incell}arcsec'
mapsize = ''
pointingspacing = '1.0arcmin'
setpointings = True
predict = True
complist = ''
refdate = '2012/06/21/03:25:00'
"""
    # Spectral setup
    txt += f"""inchan = {inchan}
incenter = '{freq:17.15e}Ghz'
inwidth = '{inwidth:17.15e}Ghz'
"""
    if simobs_custom:
        txt += "vis = '" + ms + "'\n"
    else:
        # Observing time
        txt += f"""totaltime = '{obstime}s'
integration = '{sampling_time}s'
"""

        # Configuration
        txt += f"repodir=os.getenv(\"CASAPATH\").split(\' \')[0]\n"
        if resol is None:
            if isinstance(config, str):
                txt += f"antennalist = repodir+'/data/alma/simmos/" + config + ".cfg'\n"
            elif isinstance(config, list):
                txt += f"antennalist = ["
                for i, c in enumerate(config):
                    if i > 0:
                        txt += ","
                    txt += "repodir+'/data/alma/simmos/" + c + ".cfg'"
                txt += "]\n"
        else:
            txt += f"antennalist = \"alma;"+ resol_name +"arcsec\"\n"

    # Noise
    txt += f"thermalnoise = " + th_noise + "\n"
    if is_th_noise:
        txt += f"user_pwv = {pwv}\n"
        if not simobs_custom:
            txt += "vis = project+'.noisy.ms' # clean the data with *thermal noise added*\n"

    # Imaging
    txt += f"""image = True
cleanmode = 'clark'
imsize = [{model.nx},{model.ny}]
cell = ''
niter = {n_iter}
threshold = '0.0mJy'
weighting = 'natural'
outertaper = []
stokes = 'I'
"""

    # default simalma values
    txt += f"""analyze = True
graphics = 'file'
overwrite = True
verbose = False
async = False
"""

    # Actual script
    if simobs_custom:
        if is_image:
            txt += f"mode = 'cont'\n"
        else:
            txt += f"mode = 'line'\n"
        txt += f"simobs_custom()\n"
        txt += (
            "exportfits(imagename=project+'/'+project,fitsimage='"
            + simu_name
            + f".fits',overwrite=True)\n"
        )
    else:
        txt += f"simalma()\n"
        txt += (
            "exportfits(imagename=project+'/'+project+'."
            + resol_name_script
            + f".noisy.image',fitsimage='"
            + simu_name
            + f".fits',overwrite=True)\n"
        )

    # txt += "pl.savefig('"+simu_name+f".png')\n"
    txt += "exit\n"

    # writing the script to disk
    outfile = open(workdir + simu_name + ".py", 'w')
    outfile.write(txt)
    outfile.close()

    if not only_prepare:
        return _run_CASA(simu_name)


def _run_CASA(simu_name, node_dir=""):
    print("Starting casa ...")

    workdir = "CASA/" + node_dir + "/"
    _CASA_clean(workdir)

    # -- Do we run the simulator with phase noise ?
    # fh = cfitsio_open(workdir+simu_name+".raw.fits","r");
    # phase_noise = cfitsio_get(fh, "phase_noise") ;
    # cfitsio_close,fh;

    # Running the simulator
    homedir = os.getcwd()
    os.chdir(workdir)

    # cmd="/Applications/CASA.app/./Contents/Resources/python/regressions/admin/runcasa_from_shell.sh 0 "+simu_name+".py"
    cmd = "casa --nogui -c " + simu_name + ".py"
    subprocess.call(cmd.split())
    os.chdir(homedir)

    # system, "mv "+workdir+"ALMA_disk.png  "+workdir+simu_name+".png" ;
    # system, "mv "+workdir+"ALMA_disk.fits "+workdir+simu_name+".fits" ;
    #  system, "mv "+workdir+"casapy.log "+workdir+simu_name+".log" ;

    # if (phase_noise) system, "mv "+workdir+"ALMA_disk_phase-noise.fits "+workdir+simu_name+"_phase-noise.fits" ;

    _CASA_clean(workdir)

    print("CASA simulation DONE")
    # write, "Simulation done in ",tac(), " sec" ;

    return simu_name


def _CASA_clean(workdir):
    cmd = (
        "rm -rf "
        + workdir
        + "DISK* "
        + workdir
        + "disk.fits "
        + workdir
        + "*.last "
        + workdir
        + "disk.py "
        + workdir
        + "ALMA_disk.png "
        + workdir
        + "ALMA_disk.fits "
        + workdir
        + "*.log*"
    )
    subprocess.call(cmd.split())



import os, warnings
import astropy.io.fits as fits
from astropy.convolution import Gaussian2DKernel, convolve_fft, convolve
import matplotlib.colors as mcolors
from matplotlib.patches import Ellipse
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
from scipy import interpolate
from scipy.ndimage import convolve1d
import scipy.constants as sc

try:
    import progressbar
except ImportError:
    warnings.warn("progressbar is not present", UserWarning)

from .parameters import Params, find_parameter_file
from .utils import FWHM_to_sigma, default_cmap, Wm2_to_Tb, Jy_to_Tb,  Wm2_to_Jy, add_colorbar

DEFAULT_LINE_FILE = "lines.fits.gz"

class Line:

    def __init__(self, dir=None, line_file=None, **kwargs):

        # Correct path if needed
        dir = os.path.normpath(os.path.expanduser(dir))
        self.dir = dir

        # Search for parameter file
        para_file = find_parameter_file(dir)

        # Read parameter file
        self.P = Params(para_file)

        # If user specified line_file
        if line_file is not None:
            self._line_file = line_file
        # otherwise
        else:
            self._line_file = DEFAULT_LINE_FILE

        # Read model results
        self._read(**kwargs)

    def _read(self):
        # Read ray-traced image
        try:
            hdu = fits.open(self.dir + "/" + self._line_file)
            self.lines = hdu[0].data
            # Read a few keywords in header
            self.pixelscale = hdu[0].header['CDELT2'] * 3600.0  # arcsec
            self.unit = hdu[0].header['BUNIT']
            self.cx = hdu[0].header['CRPIX1']
            self.cy = hdu[0].header['CRPIX2']
            self.nx = hdu[0].header['NAXIS1']
            self.ny = hdu[0].header['NAXIS2']
            self.nv = hdu[0].header['NAXIS3']

            if self.unit == "JY/PIXEL":
                self.is_casa = True
                self.restfreq = hdu[0].header['RESTFREQ']
                self.freq = np.array([self.restfreq])
                self.velocity_type = hdu[0].header['CTYPE3']
                if self.velocity_type == "VELO-LSR":
                    self.CRPIX3 = hdu[0].header['CRPIX3']
                    self.CRVAL3 = hdu[0].header['CRVAL3']
                    self.CDELT3 = hdu[0].header['CDELT3']
                    # velocity in km/s
                    self.velocity = self.CRVAL3 + self.CDELT3 * (np.arange(1, self.nv + 1) - self.CRPIX3)
                else:
                    raise ValueError("Velocity type is not recognised")

                self.nu = self.restfreq * (1 - self.velocity * 1000 / sc.c)

                try:
                    self.star_positions = hdu[1].data
                except:
                    self.star_positions = []
                try:
                    self.star_vr = hdu[2].data
                except:
                    self.star_vr = []
                try:
                    self.star_properties = hdu[3].data
                except:
                    self.star_properties = []
            else:
                self.is_casa = False
                self.cont = hdu[1].data

                self.ifreq = hdu[2].data
                self.freq = hdu[3].data  # frequencies of the transition
                self.velocity = hdu[4].data / 1000  # km/s
                try:
                    self.star_positions = hdu[5].data
                except:
                    self.star_positions = []
                try:
                    self.star_vr = hdu[6].data
                except:
                    self.star_vr = []

            self.dv = self.velocity[1] - self.velocity[0]

            hdu.close()
        except OSError:
            print('cannot open', self._line_file)

    def plot_map(
        self,
        i=0,
        iaz=0,
        iTrans=0,
        v=None,
        iv=None,
        insert=False,
        subtract_cont=False,
        moment=None,
        psf_FWHM=None,
        bmaj=None,
        bmin=None,
        bpa=None,
        plot_beam=None,
        beam_position=(0.125, 0.125), # fraction of plot width and height
        axes_unit="arcsec",
        conv_method=None,
        fmax=None,
        fmin=None,
        fpeak=None,
        dynamic_range=1e3,
        color_scale=None,
        colorbar=True,
        colorbar_size=10,
        cmap=None,
        ax=None,
        no_xlabel=False,
        no_ylabel=False,
        no_vlabel=False,
        no_xticks=False,
        no_yticks=False,
        vlabel_position=(0.5, 0.1), # fraction of plot width and height
        vlabel_size=8, # size in points
        title=None,
        title_size=14,
        limit=None,
        limits=None,
        Tb=False,
        Jy=False,
        mJy=False,
        per_arcsec2=False,
        per_beam=False,
        Delta_v=None,
        shift_dx=0,
        shift_dy=0,
        plot_stars=False,
        sink_particle_size=6,
        sink_particle_color="cyan",
        M0_threshold=None,
        iv_support=None,
        v_minmax = None,
        rms=0,
        interpolation=None,
        # from casa - contours
        plot_type="imshow",
        alpha=1.0,
        linewidths=None,
        zorder=None,
        origin='lower',
        levels=None,
        colors=None
    ):
        # Todo:
        # - print molecular info (eg CO J=3-2)

        if ax is None:
            ax = plt.gca()

        # -- Selecting channel corresponding to a given velocity
        if v is not None:
            iv = np.abs(self.velocity - v).argmin()
            print("Selecting channel #", iv)

        # --- Compute pixel scale and extent of image
        if axes_unit.lower() == 'arcsec':
            pix_scale = self.pixelscale
            xlabel = r'$\Delta$ RA (")'
            ylabel = r'$\Delta$ Dec (")'
            xaxis_factor = -1
        elif axes_unit.lower() == 'au':
            pix_scale = self.pixelscale * self.P.map.distance
            xlabel = 'Distance from star (au)'
            ylabel = 'Distance from star (au)'
            xaxis_factor = 1
        elif axes_unit.lower() == 'pixels' or axes_unit.lower() == 'pixel':
            pix_scale = 1
            xlabel = r'$\Delta$ x (pix)'
            ylabel = r'$\Delta$ y (pix)'
            xaxis_factor = 1
        else:
            raise ValueError("Unknown unit for axes_units: " + axes_unit)
        halfsize = np.asarray(self.lines.shape[-2:]) / 2 * pix_scale
        extent = [-halfsize[0]*xaxis_factor-shift_dx, halfsize[0]*xaxis_factor-shift_dx, -halfsize[1]-shift_dy, halfsize[1]-shift_dy]
        self.extent = extent

        # -- set color map
        if cmap is None:
            if moment in [1,9]:
                cmap = "RdBu_r"
            else:
                cmap = default_cmap
        unit = self.unit

        # -- beam or psf : psf_FWHM and bmaj and bmin are in arcsec, bpa in deg
        i_convolve = False
        beam = None
        if psf_FWHM is not None:
            # in pixels
            sigma = psf_FWHM / self.pixelscale * FWHM_to_sigma
            beam = Gaussian2DKernel(sigma)
            i_convolve = True
            bmin = psf_FWHM
            bmaj = psf_FWHM
            bpa = 0
            if plot_beam is None:
                plot_beam = True

        if bmaj is not None:
            sigma_x = bmin / self.pixelscale * FWHM_to_sigma  # in pixels
            sigma_y = bmaj / self.pixelscale * FWHM_to_sigma  # in pixels
            beam = Gaussian2DKernel(sigma_x, sigma_y, bpa * np.pi / 180)
            i_convolve = True
            if plot_beam is None:
                plot_beam = True

        # -- Selecting convolution function
        if conv_method is None:
            conv_method = convolve_fft

        # -- Selection of image to plot
        if moment is not None:
            im = self.get_moment_map(
                i=i,
                iaz=iaz,
                iTrans=iTrans,
                moment=moment,
                beam=beam,
                conv_method=conv_method,
                subtract_cont=subtract_cont,
                M0_threshold=M0_threshold,
                iv_support=iv_support,
                v_minmax=v_minmax)

            if moment in [1,2,9]:
                Tb = False

        else:
            # individual channel
            if self.is_casa:
                cube = self.lines[:, :, :]
                # im = self.lines[iv+1,:,:])
            else:
                cube = self.lines[iaz, i, iTrans, :, :, :]
                # im = self.lines[i,iaz,iTrans,iv,:,:]

                # -- continuum subtraction
                if subtract_cont:
                    cube = np.maximum(cube - self.cont[iaz, i, iTrans, np.newaxis, :, :], 0.0)

            # Convolve spectrally
            if Delta_v is not None:
                cube = self._spectral_convolve(cube, Delta_v)

            if iv is None:
                raise ValueError("plot_map needs either v or iv")
            im = cube[iv, :, :]

        # -- Convolve image
        if i_convolve:
            im = conv_method(im, beam)
            if plot_beam is None:
                plot_beam = True

        # -- Conversion to brightness temperature
        if Tb:
            if self.is_casa:
                im = Jy_to_Tb(im, self.freq[iTrans], self.pixelscale)
            else:
                im = Wm2_to_Tb(im, self.freq[iTrans], self.pixelscale)
                im = np.nan_to_num(im)
            print("Max Tb=", np.max(im), "K")

            # turning off some flags that may interfere
            per_arcsec2 = False
            per_beam = False

        # -- Conversion to Jy
        if Jy:
            if not self.is_casa:
                im = Wm2_to_Jy(im, self.freq[iTrans])
                unit = unit.replace("W.m-2", "Jy")

        # -- Conversion to mJy
        if mJy:
            if not self.is_casa:
                im = Wm2_to_Jy(im, self.freq[iTrans]) * 1e3
                unit = unit.replace("W.m-2", "mJy")

        # -- Conversion to flux per arcsec2 or per beam
        if per_arcsec2:
            im = im / self.pixelscale**2
            unit = unit.replace("pixel-1", "arcsec-2")
        if per_beam:
            beam_area = bmin * bmaj * np.pi / (4.0 * np.log(2.0))
            pix_area = self.pixelscale**2
            im *= beam_area/pix_area
            unit = unit.replace("pixel-1", "beam-1")

        # -- Adding noise
        if rms > 0.0:
            noise = np.random.randn(im.size).reshape(im.shape)
            if i_convolve:
                noise = convolve_fft(noise, beam)
            noise *= rms / np.std(noise)
            im += noise

        # --- Plot range and color map`
        _color_scale = 'lin'
        if fmax is None:
            fmax = im.max()
        if fpeak is not None:
            fmax = im.max() * fpeak
        if fmin is None:
            fmin = im.min()

        if color_scale is None:
            color_scale = _color_scale
        if color_scale == 'log':
            if fmin <= 0.0:
                fmin = fmax / dynamic_range
            norm = mcolors.LogNorm(vmin=fmin, vmax=fmax, clip=True)
        elif color_scale == 'lin':
            norm = mcolors.Normalize(vmin=fmin, vmax=fmax, clip=True)
        elif color_scale == 'sqrt':
            norm = mcolors.PowerNorm(0.5, vmin=fmin, vmax=fmax, clip=True)
        else:
            raise ValueError("Unknown color scale: " + color_scale)

        # -- Make the plot
        ax.cla()
        image = ax.imshow(im, norm=norm, extent=extent, origin='lower', cmap=cmap,interpolation=interpolation)

        if limit is not None:
            limits = [limit, -limit, -limit, limit]

        if limits is not None:
            ax.set_xlim(limits[0], limits[1])
            ax.set_ylim(limits[2], limits[3])

        if not no_xlabel:
            ax.set_xlabel(xlabel)
        if not no_ylabel:
            ax.set_ylabel(ylabel)

        if no_xticks:
            ax.get_xaxis().set_visible(False)
        if no_yticks:
            ax.get_yaxis().set_visible(False)

        if title is not None:
            ax.set_title(title,size=title_size)

        # start from casa - contours
        if plot_type=="imshow":
            image = ax.imshow(
                im,
                norm=norm,
                extent=extent,
                origin='lower',
                cmap=cmap,
                alpha=alpha,
                interpolation=interpolation,
                zorder=zorder
            )
        elif plot_type=="contourf":
            image = ax.contourf(
                im,
                extent=extent,
                origin='lower',
                levels=levels,
                cmap=cmap,
                linewidths=linewidths,
                alpha=alpha,
                zorder=zorder
            )
        elif plot_type=="contour":
            if colors is not None:
                cmap=None
            image = ax.contour(
                im,
                extent=extent,
                origin='lower',
                levels=levels,
                cmap=cmap,
                linewidths=linewidths,
                alpha=alpha,
                zorder=zorder,
                colors=colors
            )
        # end from casa - contours

        # -- Color bar
        if colorbar:
            cb = add_colorbar(image)
            formatted_unit = unit.replace("-1", "$^{-1}$").replace("-2", "$^{-2}$")
            if moment == 0:
                if Tb:
                    cb.set_label("\int T$_\mathrm{b}\,\mathrm{d}v$ (K.km.s$^{-1}$)",size=colorbar_size)
                else:
                    cb.set_label("Flux (" + formatted_unit + ".km.s$^{-1}$)",size=colorbar_size)
            elif moment == 1 or moment == 9:
                cb.set_label("Velocity (km.s$^{-1})$",size=colorbar_size)
            elif moment == 2:
                cb.set_label("Velocity dispersion (km.s$^{-1}$)",size=colorbar_size)
            else:
                if Tb:
                    cb.set_label("T$_\mathrm{b}$ (K)",size=colorbar_size)
                else:
                    cb.set_label("Flux (" + formatted_unit + ")",size=colorbar_size)

        # -- Adding velocity
        if moment is None:
            if not no_vlabel:
                vlabx, vlaby = vlabel_position
                ax.text(
                    vlabx, vlaby,
                    f"$\Delta$v={self.velocity[iv]:<4.2f}$\,$km/s",
                    horizontalalignment='center',
                    size=vlabel_size,
                    color="white",
                    transform=ax.transAxes,
                )

        # --- Adding beam
        if plot_beam:
            dx, dy = beam_position
            beam = Ellipse(
                ax.transLimits.inverted().transform((dx, dy)),
                width=bmin,
                height=bmaj,
                angle=-bpa,
                fill=True,
                color="grey",
            )
            ax.add_patch(beam)

        #-- Add stars
        if plot_stars:
            factor = pix_scale / self.pixelscale
            if isinstance(plot_stars,bool):
                x_stars = self.star_positions[0,iaz,i,:] * factor * (-xaxis_factor)
                y_stars = self.star_positions[1,iaz,i,:] * factor
            else: # int or list of int
                x_stars = self.star_positions[0,iaz,i,plot_stars] * factor * (-xaxis_factor)
                y_stars = self.star_positions[1,iaz,i,plot_stars] * factor
            ax.scatter(x_stars-shift_dx, y_stars-shift_dy,
                        color=sink_particle_color,s=sink_particle_size)

        #-- Saving the last plotted quantity
        self.last_image = im

        plt.sca(ax)

        return image

    def plot_line(
        self,
        i=0,
        iaz=0,
        iTrans=0,
        psf_FWHM=None,
        bmaj=None,
        bmin=None,
        bpa=None,
        plot_beam=False,
        plot_cont=True,
        subtract_cont=False
    ):

        if self.is_casa:
            line = np.sum(self.lines[:, :, :], axis=(1, 2))
            ylabel = "Flux (Jy)"
        else:
            if subtract_cont:
                lines = np.maximum(self.lines[iaz, i, iTrans, :, :, :] - self.cont[iaz, i, iTrans, np.newaxis, :, :], 0.0)
            else:
                lines = self.lines[iaz, i, iTrans, :, :, :]

            line = np.sum(lines, axis=(1, 2))
            ylabel = "Flux (W.m$^{-2}$)"

        print(line)
        plt.semilogy(self.velocity, line)

        if plot_cont:
            if self.is_casa:
                Fcont = 0.5 * (line[0] + line[-1])  # approx the continuum
            else:
                Fcont = np.sum(self.cont[iaz, i, iTrans, :, :])
            plt.plot([self.velocity[0], self.velocity[-1]], [Fcont, Fcont])

        xlabel = "v (m.s$^{-1}$)"

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)

    def get_moment_map(self, i=0, iaz=0, iTrans=0, moment=0,
                       beam=None, conv_method=None,subtract_cont=False,M0_threshold=None,
                       iv_support=None, v_minmax = None):
        """
        This returns the moment maps in physical units, ie:
         - M1 is the average velocity (km/s)
         - M2 is the velocity dispersion (km/s)
        """
        if self.is_casa:
            cube = np.copy(self.lines[:, :, :])
        else:
            cube = np.copy(self.lines[iaz, i, iTrans, :, :, :])

        if subtract_cont:
            cube = np.maximum(cube - self.cont[iaz, i, iTrans, np.newaxis, :, :], 0.0)

        dv = self.velocity[1] - self.velocity[0]
        v = self.velocity

        if v_minmax is not None:
            vmin = np.min(v_minmax)
            vmax = np.max(v_minmax)
            iv_support = np.array(np.where(np.logical_and((self.velocity > vmin),(self.velocity < vmax)))).ravel()
            print("Selecting channels:", iv_support)

        if iv_support is None:
            nv = self.nv
        else:
            nv = len(iv_support)
            v = v[iv_support]
            cube = cube[iv_support,:,:]

        # Moment 0, 1 and 2
        if beam is None:
            M0 = np.sum(cube, axis=0) * dv
        else:
            if moment == 0:
                M0 = np.sum(cube, axis=0) * dv
                M0 = conv_method(M0, beam)
            else:  # We need to convolve each channel indidually
                print("Convolving individual channel maps, this may take a bit of time ....")
                try:
                    bar = progressbar.ProgressBar(
                        maxval=self.nv,
                        widgets=[
                            progressbar.Bar('=', '[', ']'),
                            ' ',
                            progressbar.Percentage(),
                        ]
                    )
                    bar.start()
                except:
                    pass
                for iv in range(nv):
                    try:
                        bar.update(iv + 1)
                    except:
                        pass
                    channel = np.copy(cube[iv, :, :])
                    cube[iv, :, :] = conv_method(channel, beam)
                M0 = np.sum(cube, axis=0) * dv
                try:
                    bar.finish()
                except:
                    pass

        if moment >= 1:
            M1 = np.sum(cube[:, :, :] * v[:, np.newaxis, np.newaxis], axis=0) * dv / M0

        if moment == 2:
            M2 = np.sqrt(np.sum(cube[:, :, :]
                    * (v[:, np.newaxis, np.newaxis] - M1[np.newaxis, :, :])**2,
                    axis=0,) * dv / M0)

        # Peak flux
        if moment == 8:
            M8 = np.max(cube, axis=0)

        # Velocity of the peak
        if moment == 9:
            vmax_index = np.argmax(cube, axis=0)
            M9 = v[(vmax_index)]

            #print(vmax_index.shape)
            #
            ## Extract the maximum and neighboring pixels
            #print("test1")
            #f_max = cube[(vmax_index)]
            #print(f_max.shape)
            #print("test2")
            #f_minus = cube[(vmax_index-1)]
            #print("test3")
            #f_plus = cube[(vmax_index+1)]
            #
            ## Work out the polynomial coefficients
            #print("test4")
            #a0 = 13. * f_max / 12. - (f_plus + f_minus) / 24.
            #print("test5")
            #a1 = 0.5 * (f_plus - f_minus)
            #print("test6")
            #a2 = 0.5 * (f_plus + f_minus - 2*f_max)
            #
            ## Compute the maximum of the quadratic
            #x_max = idx - 0.5 * a1 / a2
            #y_max = a0 - 0.25 * a1**2 / a2
            #
            #M9 = xmax

        if moment == 0:
            M = M0
        elif moment == 1:
            M = M1
        elif moment == 2:
            M = M2
        elif moment == 8:
            M = M8
        elif moment == 9:
            M = M9

        if M0_threshold is not None:
            M = np.ma.masked_where(M0 < M0_threshold, M)

        return M


    def _spectral_convolve(self, cube, Delta_v):

        print("Spectral convolution at ", Delta_v, "km/s ->",Delta_v / self.dv, "channels")

        # Creating a Hanning function with k points per mcfost delta_v
        # could be optmised to have k points accrod the hanning function, but this is fast
        k = 100

        width = 2 * Delta_v / self.dv # in channels
        nchan = int(width)+1
        if nchan == 1:
            return cube
        if nchan%2 == 0:
            nchan=nchan+1

        # number of point for the high res pectral response
        n = int(width*k)+1
        if n%2 != 0:
            n=n+1

        w = np.hanning(n)
        w = w/np.sum(w)

        # rebinning the spectral response to the mcfost resolution (and padding with zeros as required)
        n_pad = (nchan * k - n) // 2

        if n_pad > 0:
            pad = np.zeros(n_pad) * 1.0
            w = np.concatenate((pad,w,pad))

        w = w.reshape(nchan,k).sum(-1)
        cube = convolve1d(cube, w, mode='constant',axis=0, cval=0.)

        return cube



from pathlib import Path
import os
from copy import copy

import pytest
from pymcfost import Params

test_dir = Path(__file__).parent

output_dir = test_dir / "artifacts"
input_dir = test_dir / "corpus"
corpus = list(input_dir.glob("*.para"))


@pytest.mark.parametrize("parafile", corpus)
def test_io_copy(parafile):
    """Read a valid .para file, check that it can still be read and written with no data corruption"""
    os.makedirs(output_dir, exist_ok=True)
    p1 = Params(str(parafile))
    s1 = copy(str(p1))

    output_file = str(output_dir / "_".join(["copy", parafile.name]))
    p1.writeto(output_file)
    p2 = Params(output_file)
    s2 = str(p2)

    s3 = str(p1)
    assert s3 == s1
    assert s2 == s1


@pytest.mark.parametrize(
    "parafile,expected",
    [
        ("ref3.0.para", (False, True, True, True)),
        ("ref3.0_switch_bools.para", (True, False, False, False)),
    ],
)
def test_bool_parsing(parafile, expected):
    p = Params(input_dir / parafile)
    result = (
        p.simu.radial_migration,
        p.simu.image_symmetry,
        p.mol.compute_pop,  # lpop
        p.stars[0].is_bb,
    )
    assert result == expected



import os, warnings
import astropy.io.fits as fits
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import numpy as np

try:
    import mpl_scatter_density
except ImportError:
    warnings.warn("mpl_scatter_density is not present", UserWarning)

from .parameters import Params, find_parameter_file
from .disc_structure import _plot_cutz, check_grid
from .utils import DustExtinction
from .run import run


class SED:

    _sed_th_file = ".sed_th.fits.gz"
    _sed_mc_file = "sed_mc.fits.gz"
    _sed_rt_file = "sed_rt.fits.gz"
    _temperature_file = "Temperature.fits.gz"
    _nlte_temperature_file = "Temperature_nLTE.fits.gz"

    def __init__(self, dir=None, **kwargs):
        # Correct path if needed
        dir = os.path.normpath(os.path.expanduser(dir))
        if dir[-7:] != "data_th":
            dir = os.path.join(dir, "data_th")
        self.dir = dir
        self.basedir = dir[:-8]

        # Search for parameter file
        para_file = find_parameter_file(dir)

        # Read parameter file
        self.P = Params(para_file)

        # Read model results
        self._read(**kwargs)


    def _read(self):
        # Read SED files
        try:
            hdu = fits.open(self.dir + "/" + self._sed_th_file)
            self._sed_th = hdu[0].data
            self._wl_th = hdu[1].data
            hdu.close()
        except OSError:
            print('cannot open', self._sed_th_file)

        try:
            hdu = fits.open(self.dir + "/" + self._sed_mc_file)
            self._sed_mc = hdu[0].data
            hdu.close()
        except OSError:
            print('cannot open', self._sed_mc_file)

        try:
            hdu = fits.open(self.dir + "/" + self._sed_rt_file)
            self.sed = hdu[0].data
            self.wl = hdu[1].data
            hdu.close()
        except OSError:
            print('cannot open', self._sed_rt_file)

        # Read temperature file
        try:
            hdu = fits.open(self.dir + "/" + self._temperature_file)
            self.T = hdu[0].data
            hdu.close()
        except OSError:
            print('Warning: cannot open', self._temperature_file)
            try:
                hdu = fits.open(self.dir + "/" + self._nlte_temperature_file)
                self.T = hdu[0].data
                hdu.close()
            except OSError:
                print('Warning: cannot open', self._temperature_file)


    def plot(self, i, iaz=0, MC=False, contrib=False, Av=0, Rv=3.1, color="black", **kwargs):

        # extinction
        if Av > 0:
            ext = DustExtinction(Rv=Rv)
            redenning = ext.redenning(self.wl, Av)
        else:
            redenning = 1.0

        if MC:
            sed = self._sed_mc[:, iaz, i, :] * redenning
        else:
            sed = self.sed[:, iaz, i, :] * redenning

        plt.loglog(self.wl, sed[0, :], color=color, **kwargs)
        plt.xlabel("$\lambda$ [$\mu$m]")
        plt.ylabel("$\lambda$.F$_{\lambda}$ [W.m$^{-2}$]")

        if contrib:
            n_type_flux = sed.shape[0]
            if n_type_flux in [5, 8, 9]:
                if n_type_flux > 5:
                    n_pola = 4
                else:
                    n_pola = 1
                linewidth = 0.5
                plt.loglog(
                    self.wl, sed[n_pola, :], linewidth=linewidth, color="magenta"
                )
                plt.loglog(
                    self.wl, sed[n_pola + 1, :], linewidth=linewidth, color="blue"
                )
                plt.loglog(
                    self.wl, sed[n_pola + 2, :], linewidth=linewidth, color="red"
                )
                plt.loglog(
                    self.wl, sed[n_pola + 3, :], linewidth=linewidth, color="green"
                )
            else:
                ValueError("There is no contribution data")

    def verif(self):
        # Compares the SED from step 1 and step 2 to check if energy is properly conserved
        current_fig = plt.gcf().number
        plt.figure(20)
        plt.loglog(self._wl_th, np.sum(self._sed_th[0, :, :], axis=0), color="black")
        plt.loglog(self.wl, np.sum(self._sed_mc[0, 0, :, :], axis=0), color="red")
        plt.figure(current_fig)

    def plot_T(self, iaz=0, log=False, Tmin=None, Tmax=None):
        # For a cylindrical or spherical grid only at the moment
        # Todo:
        #  - automatically compute call mcfost to compute the grid
        # as required
        #  - add functions for radial and vertical cuts
        # We test if the grid structure already exist, if not we try to read it

        grid = check_grid(self)

        plt.cla()

        if grid.ndim > 2:
            Voronoi = False

            r = grid[0, iaz, :, :]
            z = grid[1, iaz, :, :]
            if self.T.ndim > 2:
                T = self.T[iaz, :, :]
            else:
                T = self.T[:, :]
        else:
            Voronoi = True
            T = self.T
            r = np.sqrt(grid[0, :] ** 2 + grid[1, :] ** 2)
            ou = r > 1e-6  # Removing star
            T = T[ou]
            r = r[ou]
            z = grid[2, ou]

        if Tmin is None:
            Tmin = T.min()
        if Tmax is None:
            Tmax = T.max()

        if log:
            if Voronoi:
                # plt.scatter(r,z/r,c=T,s=0.1, norm=mcolors.LogNorm(vmin=Tmin, vmax=Tmax))
                fig = plt.gcf()
                ax = fig.add_subplot(1, 1, 1, projection='scatter_density')
                density = ax.scatter_density(
                    r,
                    z / r,
                    c=T,
                    cmap=plt.cm.RdYlBu_r,
                    dpi=100,
                    norm=mcolors.LogNorm(vmin=Tmin, vmax=Tmax),
                )
                fig.colorbar(density, label="T [K]")
            else:
                plt.pcolormesh(r, z / r, T, norm=mcolors.LogNorm(vmin=Tmin, vmax=Tmax))
                cb = plt.colorbar()
                cb.set_label('T [K]')
            plt.xscale('log')
            plt.xlabel("r [au]")
            plt.ylabel("z/r")
        else:
            if Voronoi:
                # plt.scatter(r,z,c=T,s=0.1, norm=mcolors.LogNorm(vmin=Tmin, vmax=Tmax))
                fig = plt.gcf()
                ax = fig.add_subplot(1, 1, 1, projection='scatter_density')
                density = ax.scatter_density(
                    r,
                    z,
                    c=T,
                    cmap=plt.cm.RdYlBu_r,
                    dpi=100,
                    norm=mcolors.LogNorm(vmin=Tmin, vmax=Tmax),
                )
                fig.colorbar(density, label="T [K]")
            else:
                plt.pcolormesh(r, z, T, norm=mcolors.LogNorm(vmin=Tmin, vmax=Tmax))
                cb = plt.colorbar()
                cb.set_label('T [K]')
            plt.xlabel("r [au]")
            plt.ylabel("z [au]")


    def plot_Tz(self,r=100.0, dr=5.0, log=False, **kwargs):

        _plot_cutz(self,self.T,r=r,dr=dr,log=log, **kwargs)
        plt.ylabel("T [K]")


    def plot_Tr(self, h_r=0.05, iaz=0, log=True, **kwargs):

        grid = check_grid(self)

        if grid.ndim > 2:

            if self.T.ndim > 2:
                r_mcfost = grid[0, 0, self.P.grid.nz, :]
                T = self.T[iaz, self.P.grid.nz, :]
            else:
                r_mcfost = grid[0, 0, 0, :]
                T = self.T[0, :]

            if log:
                plt.loglog(r_mcfost, T, **kwargs)
            else:
                plt.plot(r_mcfost, T, **kwargs)
        else:
            r_mcfost = np.sqrt(grid[0, :] ** 2 + grid[1, :] ** 2)
            ou = r_mcfost > 1e-6  # Removing star
            T = self.T[ou]
            r_mcfost = r_mcfost[ou]
            z_mcfost = grid[2, ou]

            # Selecting data points
            ou = abs(z_mcfost) / r_mcfost < h_r

            r_mcfost = r_mcfost[ou]
            T = T[ou]

            fig = plt.gcf()
            ax = fig.add_subplot(1, 1, 1, projection='scatter_density')
            density = ax.scatter_density(r_mcfost,T, **kwargs)

        plt.xlabel("r [au]")
        plt.ylabel("T [K]")


        return r_mcfost, T

    def spectral_index(self, wl1, wl2, i=0, iaz=0):
        pass



import copy
import os

import astropy.io.fits as fits
from astropy.convolution import Gaussian2DKernel, convolve, convolve_fft, AiryDisk2DKernel
import matplotlib.cm as cm
import matplotlib.colors as mcolors
from matplotlib.patches import Ellipse
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
import scipy.constants as sc

from .parameters import Params, find_parameter_file
from .utils import *

class Image:

    _RT_file = "RT.fits.gz"
    _MC_file = "MC.fits.gz"

    def __init__(self, dir=None, **kwargs):
        # Correct path if needed
        dir = os.path.normpath(os.path.expanduser(dir))
        self.dir = dir

        # Search for parameter file
        para_file = find_parameter_file(dir)

        # Read parameter file
        self.P = Params(para_file)

        # Read model results
        self._read(**kwargs)

    def _read(self):
        # Read ray-traced image
        try:
            hdu = fits.open(self.dir + "/" + self._RT_file)
            self.image = hdu[0].data
            # Read a few keywords in header
            self.header = hdu[0].header
            self.pixelscale = hdu[0].header['CDELT2'] * 3600.0  # arcsec
            self.unit = hdu[0].header['BUNIT']
            self.wl = hdu[0].header['WAVE']  # micron
            self.freq = sc.c / (self.wl * 1e-6)
            self.cx = hdu[0].header['CRPIX1']
            self.cy = hdu[0].header['CRPIX2']
            self.nx = hdu[0].header['NAXIS1']
            self.ny = hdu[0].header['NAXIS2']
            self.is_casa = self.unit == "JY/PIXEL"
            try:
                self.star_positions = hdu[1].data
            except:
                self.star_positions = []
            try:
                self.star_vr = hdu[2].data
            except:
                self.star_vr = []
            try:
                self.star_properties = hdu[3].data
            except:
                self.star_properties = []
            hdu.close()
        except OSError:
            print('cannot open', self._RT_file)

    def plot(
        self,
        i=0,
        iaz=0,
        vmin=None,
        vmax=None,
        dynamic_range=1e6,
        fpeak=None,
        axes_unit='arcsec',
        colorbar=True,
        colorbar_size=10,
        type='I',
        scale=None,
        pola_vector=False,
        vector_color="white",
        nbin=5,
        psf_FWHM=None,
        bmaj=None,
        bmin=None,
        bpa=None,
        plot_beam=None,
        beam_position=(0.125, 0.125), # fraction of plot width and height
        conv_method=None,
        mask=None,
        cmap=None,
        ax=None,
        no_xlabel=False,
        no_ylabel=False,
        no_xticks=False,
        no_yticks=False,
        title=None,
        limit=None,
        limits=None,
        coronagraph=None,
        rescale_r2 = False,
        clear=False,
        Tb=False,
        telescope_diameter=None,
        Jy=False,
        mJy=False,
        muJy=False,
        per_arcsec2=False,
        per_beam=False,
        shift_dx=0,
        shift_dy=0,
        plot_stars=False,
        sink_particle_size=6,
        sink_particle_color="cyan",
        norm=False,
        interpolation=None
    ):
        # Todo:
        #  - plot a selected contribution
        #  - add a mask on the star ?

        # bmin and bamj in arcsec

        if clear:
            plt.clf()

        if ax is None:
            ax = plt.gca()
            ax.cla()

        pola_needed = type in ['Q', 'U', 'Qphi', 'Uphi', 'P', 'PI', 'PA'] or pola_vector
        contrib_needed = type in ['star', 'scatt', 'em_th', 'scatt_em_th']

        if pola_needed and contrib_needed:
            raise ValueError('Cannot separate both polarisation and contributions')

        # --- We first check if the requested image is present in the mcfost fits file
        ntype_flux = self.image.shape[0]
        if ntype_flux not in (4, 8):  # there is no pola
            if pola_needed:
                raise ValueError('The model does not have polarisation data')
            n_pola = 1
        else:
            n_pola = 4
        if ntype_flux not in (5, 8):  # there is no contribution
            if contrib_needed:
                raise ValueError('The model does not have contribution data')


        # --- Compute pixel scale and extent of image
        if axes_unit.lower() == 'arcsec':
            pix_scale = self.pixelscale
            xlabel = r'$\Delta$ RA (")'
            ylabel = r'$\Delta$ Dec (")'
            xaxis_factor = -1
        elif axes_unit.lower() == 'au':
            pix_scale = self.pixelscale * self.P.map.distance
            xlabel = 'Distance from star (au)'
            ylabel = 'Distance from star (au)'
            xaxis_factor = 1
        elif axes_unit.lower() == 'pixels' or axes_unit.lower() == 'pixel':
            pix_scale = 1
            xlabel = r'$\Delta$ x (pix)'
            ylabel = r'$\Delta$ y (pix)'
            xaxis_factor = 1
        else:
            raise ValueError("Unknown unit for axes_units: " + axes_unit)
        halfsize = np.asarray(self.image.shape[-2:]) / 2 * pix_scale
        extent = [-halfsize[0]*xaxis_factor-shift_dx, halfsize[0]*xaxis_factor-shift_dx, -halfsize[1]-shift_dy, halfsize[1]-shift_dy]
        self.extent = extent

        # --- Beam or psf: psf_FWHM and bmaj and bmin are in arcsec, bpa in deg
        i_convolve = False
        beam = None

        if telescope_diameter is not None:
            # sigma of Gaussian is ~ 0.42 lambda/D
            psf_FWHM = 0.42 * self.wl*1e-6 /telescope_diameter * sigma_to_FWHM / arcsec
            print("psf FWHM is ",psf_FWHM,'"')

        if psf_FWHM is not None:
            #print("test")
            # sigma in pixels
            sigma = psf_FWHM / (self.pixelscale * 2*np.sqrt(2*np.log(2)))
            beam = Gaussian2DKernel(sigma,x_size=int(15*sigma),y_size=int(15*sigma))
            i_convolve = True
            bmin = psf_FWHM
            bmaj = psf_FWHM
            bpa = 0
            if plot_beam is None:
                plot_beam = True

        if bmaj is not None:
            sigma_x = bmin / self.pixelscale * FWHM_to_sigma  # in pixels
            sigma_y = bmaj / self.pixelscale * FWHM_to_sigma  # in pixels
            beam = Gaussian2DKernel(sigma_x, sigma_y, bpa * np.pi / 180)
            i_convolve = True
            if plot_beam is None:
                plot_beam = True

        # --- Selecting convolution function
        if conv_method is None:
            conv_method = convolve_fft

        # --- Intermediate images
        if pola_needed:
            I = self.image[0, iaz, i, :, :]
            Q = self.image[1, iaz, i, :, :]
            U = self.image[2, iaz, i, :, :]
        elif contrib_needed:
            if type == "star":
                I = self.image[n_pola, iaz, i, :, :]
            elif type == "scatt":
                I = self.image[n_pola + 1, iaz, i, :, :]
            elif type == "em_th":
                I = self.image[n_pola + 2, iaz, i, :, :]
            elif type == "scatt_em_th":
                I = self.image[n_pola + 3, iaz, i, :, :]
        else:
            if self.is_casa:
                I = self.image[iaz, i, :, :]
            else:
                I = self.image[0, iaz, i, :, :]

        # --- Convolution with psf
        if i_convolve:
            I = conv_method(I, beam)
            if pola_needed:
                Q = conv_method(Q, beam)
                U = conv_method(U, beam)

        # -- Conversion to brightness temperature
        if Tb:
            if self.is_casa:
                I = Jy_to_Tb(I, self.freq, self.pixelscale)
            else:
                I = Wm2_to_Tb(I, self.freq, self.pixelscale)
                I = np.nan_to_num(I)
                print("Max Tb=", np.max(I), "K")


        # -- Conversion to Jy
        if Jy:
            if not self.is_casa:
                I = Wm2_to_Jy(I, self.freq)

        # -- Conversion to mJy
        if mJy:
            if not self.is_casa:
                I = Wm2_to_Jy(I, self.freq) * 1e3
            else:
                I *= 1e6

        # -- Conversion to microJy
        if muJy:
            if not self.is_casa:
                I = Wm2_to_Jy(I, self.freq) * 1e6
            else:
                I *= 1e6

        # --- Coronagraph: in mas
        if coronagraph is not None:
            halfsize = np.asarray(self.image.shape[-2:]) / 2
            posx = np.linspace(-halfsize[0]+shift_dx/pix_scale,
                                halfsize[0]+shift_dx/pix_scale, self.nx)
            posy = np.linspace(-halfsize[1]-shift_dy/pix_scale,
                                halfsize[1]-shift_dy/pix_scale, self.ny)
            meshx, meshy = np.meshgrid(posx, posy)
            radius_pixel = np.sqrt(meshx ** 2 + meshy ** 2)
            radius_mas = radius_pixel * pix_scale * 1000
            I[radius_mas < coronagraph] = 0.0
            if pola_needed:
                Q[radius_mas < coronagraph] = 0.0
                U[radius_mas < coronagraph] = 0.0

        # --- Selecting image to plot
        unit = self.unit
        flux_name = type
        if type in ('I','star','scatt','em_th','scatt_em_th'):
            flux_name = 'Flux density'
            im = I
            _scale = 'log'

        elif type == 'Q':
            im = Q
            _scale = 'symlog'
        elif type == 'U':
            im = U
            _scale = 'symlog'
        elif type == 'P':
            I = I + (I == 0.0) * 1e-30
            im = 100 * np.sqrt((Q / I) ** 2 + (U / I) ** 2)
            unit = "%"
            _scale = 'lin'
        elif type == 'PI':
            im = np.sqrt(np.float64(Q) ** 2 + np.float64(U) ** 2)
            _scale = 'log'
        elif type in ('Qphi', 'Uphi'):
            X = np.arange(1, self.nx + 1) - self.cx
            Y = np.arange(1, self.ny + 1) - self.cy
            X, Y = np.meshgrid(X, Y)
            two_phi = 2 * np.arctan2(Y, X)
            if type == 'Qphi':
                im = Q * np.cos(two_phi) + U * np.sin(two_phi)
            else:  # Uphi
                im = -Q * np.sin(two_phi) + U * np.cos(two_phi)
            _scale = 'symlog'

         # --- Rescale image by r^2
        if rescale_r2:
            halfsize = np.asarray(self.image.shape[-2:]) / 2
            posx = np.linspace(-halfsize[0]+shift_dx/pix_scale,
                                halfsize[0]+shift_dx/pix_scale, self.nx)
            posy = np.linspace(-halfsize[1]-shift_dy/pix_scale,
                                halfsize[1]-shift_dy/pix_scale, self.ny)
            meshx, meshy = np.meshgrid(posx, posy)
            radius_pixel2 = meshx**2 + meshy**2
            im = im * radius_pixel2
            # Normalizing to 1, as units become meaningless
            # Computing max of plotted region and normalize to it
            if limit is not None:
                limits = [limit, -limit, -limit, limit]
            if limits is not None:
                limit_pix_xmin = int(halfsize[0]+(-shift_dx-limits[0])/pix_scale)
                limit_pix_xmax = int(halfsize[0]+(-shift_dx-limits[1])/pix_scale)
                limit_pix_ymin = int(halfsize[1]+(shift_dy+limits[2])/pix_scale)
                limit_pix_ymax = int(halfsize[1]+(shift_dy+limits[3])/pix_scale)
                im = im/np.max(I[limit_pix_xmin:limit_pix_xmax,
                                limit_pix_ymin:limit_pix_ymax])
            else:
                im = im/np.max(im) # Normalizing to 1 over entire image

        if Tb:
            flux_name = "Tb"
            _scale = "lin"
            unit = "K"
            # turning off some flags that may interfere
            per_arcsec2 = False
            per_beam = False

        if Jy:
            unit = "Jy.pixel-1"

        if mJy:
            unit = "mJy.pixel-1"

        if rescale_r2:
            unit = "arbitrary units" # max == 1

        # -- Conversion to flux per arcsec2 or per beam
        if per_arcsec2:
            im = im / self.pixelscale**2
            unit = unit.replace("pixel-1", "arcsec-2")

        if per_beam:
            beam_area = bmin * bmaj * np.pi / (4.0 * np.log(2.0))
            pix_area = self.pixelscale**2
            im *= beam_area/pix_area
            unit = unit.replace("pixel-1", "beam-1")

        if norm:
            im = im / np.max(im)
            flux_name = "Normalised flux"
            unit = ""

        # --- Plot range and color scale
        if vmax is None:
            vmax = im.max()
        if fpeak is not None:
            vmax = im.max() * fpeak
        if vmin is None:
            if type in ["Q", "U"]:
                vmin = -vmax
            else:
                vmin = 1e-3 * vmax

        if scale is None:
            scale = _scale
        if scale == 'symlog':
            norm = mcolors.SymLogNorm(1e-6 * vmax, vmin=vmin, vmax=vmax, clip=True)
        elif scale == 'log':
            if vmin <= 0.0:
                vmin = 1e-5 * vmax
            if vmin < 0.9e-5 * vmax:
                print("WARNING : vmin ~< 1e-6 vmax may crash with recent versions of matplotlib")
            norm = mcolors.LogNorm(vmin=vmin, vmax=vmax, clip=True)
        elif scale == 'lin':
            norm = mcolors.Normalize(vmin=vmin, vmax=vmax, clip=True)
        elif color_scale == 'sqrt':
            norm = mcolors.PowerNorm(0.5, vmin=vmin, vmax=vmax, clip=True)
        else:
            raise ValueError("Unknown color scale: " + scale)

        # --- Set color map
        if cmap is None:
            cmap = default_cmap
        try:
            cmap = copy.copy(cm.get_cmap(cmap))
        except:
            raise ValueError("Unknown colormap: " + cmap)
        try:
            cmap.set_bad(cmap.colors[0])
        except:
            try:
                cmap.set_bad(cmap(0.0))
            except:
                raise Warning("Can't set bad values from given colormap")

        # --- Making the actual plot
        image = ax.imshow(im, norm=norm, extent=extent, origin='lower', cmap=cmap, interpolation=interpolation)

        if limit is not None:
            limits = [limit, -limit, -limit, limit]

        if limits is not None:
            ax.set_xlim(limits[0], limits[1])
            ax.set_ylim(limits[2], limits[3])

        if not no_xlabel:
            ax.set_xlabel(xlabel)
        if not no_ylabel:
            ax.set_ylabel(ylabel)

        if no_xticks:
            ax.get_xaxis().set_visible(False)
        if no_yticks:
            ax.get_yaxis().set_visible(False)

        if title is not None:
            ax.set_title(title)

        # --- Colorbar
        if colorbar:
            cb = add_colorbar(image)
            formatted_unit = unit.replace("-1", "$^{-1}$").replace("-2", "$^{-2}$")
            if unit != "":
                cb.set_label(flux_name + " (" + formatted_unit + ")",size=colorbar_size)
            else:
                cb.set_label(flux_name,size=colorbar_size)


        # --- Overplotting polarisation vectors
        if pola_vector:
            X = (np.arange(1, self.nx + 1) - self.cx) * pix_scale
            Y = (np.arange(1, self.ny + 1) - self.cy) * pix_scale
            X, Y = np.meshgrid(X, Y)

            Xb = bin_image(X, nbin, func=np.mean)
            Yb = bin_image(Y, nbin, func=np.mean)
            Ib = bin_image(I, nbin)
            Qb = bin_image(Q, nbin)
            Ub = bin_image(U, nbin)

            pola = 100 * np.sqrt((Qb / np.maximum(Ib,1e-300)) ** 2 + (Ub / np.maximum(Ib,1e-300)) ** 2)
            theta = 0.5 * np.arctan2(Ub, Qb)
            # Ref is N (vertical axis) --> sin, and Est is toward left --> -
            pola_x = pola * np.sin(theta)
            pola_y = pola * np.cos(theta)

            ax.quiver(
                Xb,
                Yb,
                pola_x,
                pola_y,
                headwidth=0,
                headlength=0,
                headaxislength=0.0,
                pivot='middle',
                color=vector_color,
            )

        # --- Adding beam
        if plot_beam:
            dx, dy = beam_position
            beam = Ellipse(
                ax.transLimits.inverted().transform((dx, dy)),
                width=bmin,
                height=bmaj,
                angle=-bpa,
                fill=True,
                color="grey",
            )
            ax.add_patch(beam)

        # --- Adding mask
        if mask is not None:
            dx = 0.5
            dy = 0.5
            mask = Ellipse(
                ax.transLimits.inverted().transform((dx, dy)),
                width=2 * mask,
                height=2 * mask,
                fill=True,
                color='grey',
            )
            ax.add_patch(mask)

        #-- Add stars
        if plot_stars:
            # star_position in arcsec by default
            factor =  pix_scale / self.pixelscale
            if isinstance(plot_stars,bool):
                x_stars = self.star_positions[0,iaz,i,:] * factor * (-xaxis_factor)
                y_stars = self.star_positions[1,iaz,i,:] * factor
            else: # int or list of int
                x_stars = self.star_positions[0,iaz,i,plot_stars] * factor * (-xaxis_factor)
                y_stars = self.star_positions[1,iaz,i,plot_stars] * factor
            ax.scatter(x_stars-shift_dx, y_stars-shift_dy,
                        color=sink_particle_color,s=sink_particle_size)

        #-- Saving the last plotted quantity
        self.last_image = im

        plt.sca(ax)

        # --- Return
        return image

    def calc_vis(
        self,
        i=0,
        iaz=0,
        hor=True,
        Jy=False,
        klambda=False,
        Mlambda=False,
        color='black',
    ):

        # smoothing, christophe suggest I may need this later
        #   ou = 1;
        #   if (!is_void(champ)) {
        #     size=model.P.map.ny;
        #     x=indgen(size)(,-:1:size) - (size/2+1);
        #     y=indgen(size)(-:1:size,) - (size/2+1);
        #     distance = abs(x,y);

        #     ou = (distance * pix_size < 0.5 * champ) ;

        #     if (gauss==1) {
        #       FWHM = champ / pix_size; // OK : le FWHM est equivalent a la largeur de la porte !!! Cool !
        #       sigma = FWHM / (2*sqrt(2*log(2))); // en sec
        #       ou = gauss_kernel(size, sigma) ;
        #     }
        #     // write, "Applying a field of view of ", champ, "as" ;

        #     if (champ > 0.5 * im_size) {
        #       write, "WARNING : image seems small to aply the filed of view accurately" ;
        #       write, im_size, champ ;
        #     }
        #   }

        # error message if klambda and Mlambda sccales are selected
        if klambda and Mlambda:
            raise Exception(
                "Cannot plot visabilities on two different scales (k and M), set one to False"
            )

        # Selecting image
        im = self.image[0, iaz, i, :, :]

        # padding the image for a smoother curve
        def pad_with(vector, pad_width, iaxis, kwargs):
            pad_value = kwargs.get('padder', 0)
            vector[: pad_width[0]] = pad_value
            vector[-pad_width[1] :] = pad_value
            return vector

        im = np.pad(im, 1000, pad_with)

        # fft
        fim = np.real(np.fft.fft2(np.fft.fftshift(im)))

        # Baselines
        size = len(fim)
        center = size / 2

        # converting from arcsecond to radian
        pix_size = self.pixelscale / 3600.0 * np.pi / 180.0

        # pixel size in the uv plane
        pix_fft = 1.0 / pix_size

        # pixel in wavelength (normalising in wavelength)
        pix = self.wl * 1e-6 * pix_fft

        baselines = np.linspace(0, int(pix / 2), int(size / 2))

        # visabilities
        if hor:
            vis = fim[0, 0 : int(size / 2)]
        else:
            vis = fim[0 : int(size / 2), 0]

        # convert to Jy
        if Jy:
            Wm2_to_Jy(vis, sc.c / self.wl)
            ylabel = "Correlated flux (Jy)"
        else:
            ylabel = "Correlated flux (W.m$^{-2}$.Hz$^{-1}$)"

        if klambda:
            baselines = baselines / (self.wl * 1e-3)
            xlabel = "Baselines (k$\lambda$)"
        elif Mlambda:
            baselines = baselines / (self.wl * 1e-6)
            xlabel = "Baselines (M$\lambda$)"
        else:
            xlabel = "Baselines (m)"

        plt.plot(baselines, vis, color=color)
        plt.ylabel(ylabel)
        plt.xlabel(xlabel)

        return baselines, vis, fim


    def writeto(self, filename, **kwargs):
        fits.writeto(os.path.normpath(os.path.expanduser(filename)),self.last_image, self.header, **kwargs)

    def get_planet_rPA(self,iplanet):
        "Return the projected radius (arcsec) and PA of planet #iplanet in the image"
        dx, dy = self.star_positions[:,0,0,iplanet] - self.star_positions[:,0,0,0]
        PA = np.rad2deg(np.arctan2(dy,-dx)) - 90
        r = np.hypot(dx,dy)

        return r, PA

def spectral_index(model1, model2, i=0, iaz=0):

    log_nuFnu1 = np.log(np.maximum(model1.image[0, iaz, i, :, :], 1e-300))
    log_nuFnu2 = np.log(np.maximum(model2.image[0, iaz, i, :, :], 1e-300))

    dlog_nu = np.log(model1.wl) - np.log(model2.wl)

    return (log_nuFnu2 - log_nuFnu1) / dlog_nu - 3



import numpy as np
import astropy.io.fits as fits
import astropy.units as u
import sys
import subprocess
from .parameters import Params
from .disc_structure import Disc


def analytic_params_import(file):

    parobj = open(file)
    params = {} # this is a dictionary
    for line in parobj:
        line = line.strip() # removes additional spaces after a line and also the newline character
        if line and line[0] != '%': # the first 'if line' allows to remove blank lines.
            line = line.split()
            if len(line) == 1:
                params[line[0]] = None

            elif len(line) == 2:
                params[line[0]] = line[1]

            elif len(line) > 2:
                params[line[0]] = [ line[1] ]
                for i in range(2,len(line)):
                    params[line[0]] += [ line[i] ]

    Rmax = float(params['Rdisc'])
    if 'Rmin' in params.keys():
        Rmin = float(params['Rmin'])
    else:
        Rmin = Rdisc/50

    Nr = int(params['Nr'])
    Nphi = int(params['Nphi'])

    Rplanet = float(params['Rplanet'])
    PA = float(params['PA'])
    i = float(params['inclination'])
    PAp = float(params['PAp'])
    M = float(params['Mstar'])
    Mp = float(params['Mplanet'])

    cw = int(params['cw'])

    return Nr, Nphi, Rmin, Rmax, Rplanet, PAp, i, PA, M, Mp, cw



def analytic2mcfost(
    data_dir="/home/tom/Documents/Protoplanetary_Discs_Project/HD_163296_analytics_for_channels_grid/Analytics_output/hd163_4Mj",
    nz=50,
    analytic_params_file="/home/tom/Documents/Protoplanetary_Discs_Project/HD_163296_analytics_for_channels_grid/Analytics_output/hd163_4Mj/hd163.param",
    mcfost_ref_file="/home/tom/Documents/pymcfost/pymcfost/ref3.0_3D.para",
    mcfost_filename="mcfost_hd163_4Mj.para",
    fitsname='hd163_analytic_4Mj',
):

    # read in properties from analytics
    Nr, Nphi, Rmin, Rmax, Rplanet, PAp, i, PA, M, Mp, cw = analytic_params_import(analytic_params_file)

    # set no. of sink particles (code modifications needed for anything other than 2)
    Nsink = 2

    from astropy import constants as const
    from os import makedirs

    # -- Setup constants
    G = const.G.cgs.value
    au = const.au.cgs.value
    M_sol = const.M_sun.cgs.value

    # --- Reading data
    rho = np.load(data_dir + "/density.npy")
    vrad = 1e3*np.load(data_dir + "/vr.npy")
    vtheta = 1e3*np.load(data_dir + "/vphi.npy")

    # --- reshape
    Rho = rho.transpose()
    Vrad = vrad.transpose()
    Vtheta = vtheta.transpose()

    # --- mcfost : updating values from analytic kinks parameter file
    # --- Setting up a parameter file and creating the corresponding grid
    P = Params(mcfost_ref_file)
    P.grid.n_rad = Nr
    P.grid.n_rad_in = 1
    P.grid.nz = nz
    P.grid.n_az = Nphi

    # -- Disc geometry
    P.zones[0].Rin = Rmin
    P.zones[0].edge = 0.0
    P.zones[0].Rout = Rmax
    P.zones[0].Rc = 0.0

    # -- Don't compute SED
    P.simu.compute_SED = False

    # -- pixel grid
    P.map.nx = 1001
    P.map.ny = 1001

    # -- ROTATION DIR. NEEDS TO BE CORRECT FOR THE FOLLOWING TO WORK
    P.map.RT_az_min = -1*PAp # 45, planet angle
    P.map.RT_az_max = -1*PAp
    P.map.RT_n_az = 1

    # -- distance
    P.map.distance = 101.5

    # -- PA
    P.map.PA = PA - 90

    # -- inclination
    P.map.RT_imin = -1*i
    P.map.RT_imax = -1*i
    P.map.RT_ntheta = 1

    P.phot.nphot_T = 1.28e+07

    # -- velocity channel settings
    P.mol.molecule[0].v_max = 3.2     # max velocity
    P.mol.molecule[0].nv = 20         # number of velocity channels
    P.mol.molecule[0].n_trans = 1     # no. of lines in ray tracing

    # -- Turn off symmetries
    P.simu.image_symmetry = False
    P.simu.central_symmetry = False
    P.simu.axial_symmetry = False

    # -- Write new parameter file
    P.writeto(mcfost_filename)

    # --- Running mcfost to create the grid
    subprocess.call(["rm", "-rf", "data_disk", "data_disk_old"])
    subprocess.call(["mcfost", mcfost_filename, "-disk_struct"])

    # --- Reading mcfost grid
    mcfost_disc = Disc("./")

    # -- same dimension order as FARGO
    mcfost_z = mcfost_disc.z()

    # -- taking only half the grid (+z)
    mcfost_z = mcfost_z[:, nz:, :]
    mcfost_z = mcfost_z.transpose()  # dims are now, r, z, theta

    # -- defining H [au]
    h = 0.05 * mcfost_disc.r()[:, 0, :].transpose()

    # -- computing the 3D density structure for mcfost
    rho_mcfost = Rho[:, np.newaxis, :] * np.exp(-0.5 * (mcfost_z / h[:, np.newaxis, :]))
    vtheta_mcfost = np.repeat(Vtheta[:, np.newaxis, :], nz, axis=1)
    vrad_mcfost = np.repeat(Vrad[:, np.newaxis, :], nz, axis=1)
    vz_mcfost = np.zeros((220,50,200))

    # -- create master velocities array
    velocities = np.array([vrad_mcfost.transpose(),vtheta_mcfost.transpose(),vz_mcfost.transpose()])

    # -- setup HDUs for fits file
    primary_hdu = fits.PrimaryHDU(np.abs(rho_mcfost.transpose()))
    second_hdu = fits.ImageHDU(np.abs(rho_mcfost.transpose()))
    tertiary_hdu = fits.ImageHDU(velocities)

    # -- set header properties for mcfost
    primary_hdu.header['hierarch read_gas_velocity'] = 2
    primary_hdu.header['hierarch gas_dust_ratio'] = 100
    primary_hdu.header['hierarch read_gas_density'] = 1
    primary_hdu.header['read_n_a'] = 0

    # -- sink particles properties in header (code modifications needed for anything other than 2)
    primary_hdu.header['hierarch N_sink'] = Nsink

    # -- star properties in header
    primary_hdu.header['hierarch M_star'] = M          # star mass in M_sol
    primary_hdu.header['hierarch x_star'] = 0.         # star at origin
    primary_hdu.header['hierarch y_star'] = 0.
    primary_hdu.header['hierarch z_star'] = 0.
    primary_hdu.header['hierarch vx_star'] = 0.        # star not moving
    primary_hdu.header['hierarch vy_star'] = 0.
    primary_hdu.header['hierarch vz_star'] = 0.

    # -- planet properties in header
    primary_hdu.header['hierarch M_planet'] = Mp       # planet mass in M_jup
    primary_hdu.header['hierarch x_planet'] = Rplanet  # planet x position simply orbital radius [AU]
    primary_hdu.header['hierarch y_planet'] = 0.
    primary_hdu.header['hierarch z_planet'] = 0.
    primary_hdu.header['hierarch vx_planet'] = 0.      # planet velocity just keplerian rotation [m/s] (positive for anticlockwise, which is cw=-1)
    primary_hdu.header['hierarch vy_planet'] = np.sqrt(G*M*M_sol/(Rplanet*au)) / 100 * -cw
    primary_hdu.header['hierarch vz_planet'] = 0.

    # -- setup fits file
    hdul = fits.HDUList([primary_hdu, second_hdu, tertiary_hdu])

    # --- Write a fits file for mcfost
    if fitsname is not None:
        print("Writing ", fitsname, " to disk")
        hdul.writeto(fitsname + '.fits', overwrite=True)
    else:
        print("Fits file for mcfost was not created")

    return 'Done'



import os
import subprocess

_mcfost_bin = "mcfost"

def run(filename, options="", delete_previous=False, notebook=False, logfile=None, silent=False):

    if not isinstance(filename, str):
        raise TypeError("First argument to run must be a filename.")
    filename = os.path.normpath(os.path.expanduser(filename))
    if not os.path.exists(filename):
        raise IOError(filename+" does not exist")

    # Finding root directory
    root_dir = "."
    opts = options.split()
    for i, option in enumerate(opts):
        if option == "-root_dir":
            root_dir = opts[i+1]
            break

    if delete_previous:
        subprocess.call("rm -rf "+root_dir+"/data_*", shell = True)

    if silent:
        if logfile is None:
            logfile = "mcfost.log"

    # Saving output in a log file
    if logfile is not None:
        options+= " > "+logfile

    if not silent:
        print("pymcfost: Running mcfost ...")

    r = subprocess.run(_mcfost_bin+" "+filename+" "+options, shell = True)
    if r.returncode:
        raise OSError("mcfost did not run as expected, check mcfost's output")

    if not silent:
        print("pymcfost: Done")



import sys
import matplotlib.pyplot as plt
from .plutoTools import readVTKSpherical
import numpy as np
import copy
from scipy import interpolate
from astropy import units as u
from astropy import constants as c
import astropy.io.fits as fits
import subprocess
from .parameters import Params
from .disc_structure import Disc

def pluto2mcfost(
        filename,
        nr=100,
        nz=100,
        umass=1,
        udist = 10,
        mcfost_ref_file="/Users/cpinte/mcfost/src/ref3.0_3D.para",
        mcfost_filename="mcfost_PLUTO.para",
        fitsname=None,
):

    # umass in solar masses
    # udist in au

    # Inner and outer radii over which we plot the solution
    rin=1.0
    rout=3.0


    ##########################
    ## Units options
    #########################

    # Mass of central object (modifiable)
    Mstar=umass * c.M_sun
    # unit of length (=inner radius of the run, here taken to be 10 AU) (modifiable)
    R0=(udist*u.au).to(u.m)
    #Assumed surface density at R0 (eq 5 in Riols 2020)
    Sigma0=200*(R0.to(u.au)/u.au)**(-0.5)*u.g/(u.m)**2
    # Disc aspect-ratio
    epsilon=0.05

    ###########################
    # Computed quantities
    ###########################

    #Time unit
    T0=np.sqrt(R0**3/(Mstar*c.G)).to(u.s)
    #Velocity unit
    V0=R0/T0

    #density unit rho=sigma/(sqr(Zpi)H)
    rho0=Sigma0/(np.sqrt(2*np.pi)*epsilon*R0)
    # pressure unit
    P0=rho0*V0*V0

    #density unit rho=sigma/(sqr(Zpi)H)
    rho0=Sigma0/(np.sqrt(2*np.pi)*epsilon*R0)

    # Read data file
    V=readVTKSpherical(filename)

    # Apply units
    V.data['rho']= np.squeeze((rho0*V.data['rho']).value)
    V.data['prs']= np.squeeze((P0*V.data['prs']).value)
    V.data['vx1']= np.squeeze((V0*V.data['vx1']).value)
    V.data['vx2']= np.squeeze((V0*V.data['vx2']).value)
    V.data['vx3']= np.squeeze((V0*V.data['vx3']).value)

    V.r = V.r * udist # V.r in au

    # Compute the sound speed
    V.data['cs']=np.sqrt(V.data['prs']/V.data['rho'])

    n_r = V.r.size
    n_theta = V.theta.size

    # define cylindrical coordinates of the pluto model
    [r,theta] = np.meshgrid(V.r,V.theta,indexing='ij')

    rcyl = r*np.sin(theta)
    z    = r*np.cos(theta)

    # Density and velocity data
    rho = V.data['rho']

    # Create the cartesian components (in units of sound speed)
    vr = V.data['vx1']*np.sin(theta)+V.data['vx2']*np.cos(theta)
    vz = V.data['vx1']*np.cos(theta)-V.data['vx2']*np.sin(theta)
    vphi = V.data['vx3']

    # --- mcfost : updating values from PLUTO
    # --- Setting up a parameter file and creating the corresponding grid
    # Setting up the mcfost model
    P = Params(mcfost_ref_file)
    P.grid.n_rad = nr
    P.grid.n_rad_in = 1
    P.grid.nz = nz
    P.grid.n_az = 1

    P.zones[0].Rin = np.min(V.r)
    P.zones[0].edge = 0.0
    P.zones[0].Rout = np.max(V.r)
    P.zones[0].Rc = 0.0

    # Don't compute SED
    P.simu.compute_SED = False

    # -- Turn off symmetries
    P.simu.image_symmetry = False
    P.simu.central_symmetry = False
    P.simu.axial_symmetry = False

    # -- Write new parameter file
    P.writeto(mcfost_filename)

    # --- Running mcfost to create the grid
    subprocess.call(["rm", "-rf", "data_disk", "data_disk_old"])
    result = subprocess.call(["mcfost", mcfost_filename, "-disk_struct"])

    # --- Reading mcfost grid
    mcfost_disc = Disc("./")

    # --- print the mcfost radial grid to check that it is the same as FARGO's
    print("MCFOST radii=")
    print(mcfost_disc.r()[0, 0, :])

    mcfost_r = mcfost_disc.r().transpose() # dims are now, r, z, theta
    mcfost_z = mcfost_disc.z().transpose()

    # -- computing the 3D density structure for mcfost
    rho_mcfost    = interpolate.griddata((rcyl.flatten(),z.flatten()),rho.flatten() ,(mcfost_r,mcfost_z),fill_value=0)
    vr_mcfost     = interpolate.griddata((rcyl.flatten(),z.flatten()),vr.flatten()  ,(mcfost_r,mcfost_z),fill_value=0)
    vz_mcfost     = interpolate.griddata((rcyl.flatten(),z.flatten()),vz.flatten()  ,(mcfost_r,mcfost_z),fill_value=0)
    vphi_mcfost   = interpolate.griddata((rcyl.flatten(),z.flatten()),vphi.flatten(),(mcfost_r,mcfost_z),fill_value=0)

    velocities = np.array([vr_mcfost.transpose(),vphi_mcfost.transpose(),vz_mcfost.transpose()])
    #velocities = np.moveaxis(velocities, 0, -1)

    primary_hdu = fits.PrimaryHDU(np.abs(rho_mcfost.transpose()))
    second_hdu = fits.ImageHDU(np.abs(rho_mcfost.transpose()))
    tertiary_hdu = fits.ImageHDU(velocities)
    primary_hdu.header['hierarch read_gas_velocity'] = 2
    primary_hdu.header['hierarch gas_dust_ratio'] = 100
    primary_hdu.header['hierarch read_gas_density'] = 1
    primary_hdu.header['read_n_a'] = 0
    hdul = fits.HDUList([primary_hdu, second_hdu, tertiary_hdu])

    # --- Write a fits file for mcfost
    if fitsname is not None:
        print("Writing ", fitsname, " to disk")
        hdul.writeto(fitsname, overwrite=True)
    else:
        print("Fits file for mcfost was not created")

    return mcfost_r, mcfost_z, rho_mcfost, vr_mcfost, vphi_mcfost, vz_mcfost



import numpy as np
import matplotlib.pyplot as plt
from .utils import rotate_coords,rotate_to_obs_plane

def get_wake_cartesian(rp,phip,npts,rmin,rmax,HonR,q):
    '''
       planet wake formula from Rafikov (2002)
    '''
    rplot = np.linspace(rmin,rmax,npts)
    rr = rplot/rp
    phi = phip + np.sign(rmax-rmin)*(1./HonR)*((rr**(q-0.5))/(q-0.5) \
                          - (rr**(q+1.))/(q+1.) \
                          - 3./((2.*q-1.)*(q+1.)))
    xx = rplot*np.cos(phi)
    yy = rplot*np.sin(phi)
    return xx,yy

def plot_wake(xy_obs,inc,PA,HonR,q,color="black"):
    '''
       plot planet wake
       and rotate to the observational viewpoint
    '''
    inc = np.deg2rad(inc)
    PA = np.deg2rad(PA)

    # we give planet location in the observational plane
    # bad attempt to provide the location in the rotated plane
    # by a simple linear scaling (does not get the angle correct)
    x_scale, y_scale, dum = rotate_coords(xy_obs[0],xy_obs[1],0.,inc,PA)
    x_p = xy_obs[0]*(xy_obs[0]/x_scale)
    y_p = xy_obs[1]*(xy_obs[1]/y_scale)

    # planet location in the unrotated plane
    rp = np.sqrt(x_p**2 + y_p**2)
    phip = np.arctan2(y_p,x_p)
    print("rp = ",rp," phi planet = ",phip*180./np.pi)

    # radial range over which to plot the wake
    rmin = 1.e-3
    rmax = 3.*rp
    npts = 1000

    # outer wake
    xx,yy = get_wake_cartesian(rp,phip,npts,rp,rmax,HonR,q)
    xp,yp = rotate_to_obs_plane(xx,yy,inc,PA)
    plt.plot(xx,yy,color=color)

    # inner wake
    xx,yy = get_wake_cartesian(rp,phip,npts,rp,rmin,HonR,q)
    xp,yp = rotate_to_obs_plane(xx,yy,inc,PA)
    plt.plot(xx,yy,color=color)



__version__ = "0.1"
__author__ = "Christophe Pinte"

from .parameters import *
from .disc_structure import *
from .SED import *
from .image import *
from .line import *
from .fargo2mcfost import *
from .pluto2mcfost import *
from .CASA_simdata import *
from .run import *
from .wake import *
from .dust import *



import pytest
from pymcfost.parameters import _word_to_bool


true_strings = ["True", ".True.", "TRUE", ".TRUE.", "true", ".true.", "t", ".t.", "T", ".T."]
@pytest.mark.parametrize("string", true_strings)
def test_read_true_string(string):
    assert _word_to_bool(string)

false_strings = ["False", ".False.", "FALSE", ".FALSE.", "false", ".false.", "f", ".f.", "F", ".F."]
@pytest.mark.parametrize("string", false_strings)
def test_read_false_string(string):
    assert not _word_to_bool(string)



import matplotlib.pyplot as plt
from matplotlib.testing.compare import compare_images
import pymcfost as mcfost
from pymcfost.disc_structure import Disc
import pytest

import os
working_dir = os.getcwd()

data_dir = working_dir + "/tests/test_ref4.0"
out_path = working_dir + '/tests/actual'
expected = working_dir + '/tests/expected_figures'

if os.path.exists(expected) == False:
    os.system('wget http://ipag.osug.fr/public/pintec/pymcfost/testdata.tar.gz')
    os.system('tar -xvf testdata.tar.gz -C ./tests/')

if os.path.exists(out_path) == False: os.mkdir(out_path)

def data_th_plot_T(path):
    # Plot figure test_data_th_plot_T
    model = mcfost.SED(data_dir + "/data_th/")
    model.disc = Disc(model.basedir)
    fig, ax = plt.subplots()
    model.plot_T()
    ax.axis('off')
    fig.savefig('%s/test_data_th_plot_T.png' % path)


def data_th_plot_T_log(path):
    # Plot figure test_data_th_plot_T_log
    model = mcfost.SED(data_dir + "/data_th/")
    model.disc = Disc(model.basedir)
    fig, ax = plt.subplots()
    model.plot_T(log=True)
    ax.axis('off')
    fig.savefig('%s/test_data_th_plot_T_log.png' % path)

def data_th_plot_0(path):
    # Plot figure test_data_th_plot_0
    model = mcfost.SED(data_dir + "/data_th/")
    model.disc = Disc(model.basedir)
    fig, ax = plt.subplots()
    model.plot(0)
    ax.axis('off')
    fig.savefig('%s/test_data_th_plot_0.png' % path)

def data_th_plot_0_contrib(path):
    # Plot figure test_data_th_plot_0_contrib
    model = mcfost.SED(data_dir + "/data_th/")
    fig, ax = plt.subplots()
    model.plot(0, contrib=True)
    ax.set_ylim(1e-16,2e-12)   # we reduce the range on the y axis
    ax.axis('off')
    fig.savefig('%s/test_data_th_plot_0_contrib.png' % path)

def data_1_0_scattered_light(path):
    # Plot figure test_data_1_0_scattered_light
    image_1mum = mcfost.Image(data_dir + "/data_1.0/")
    fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(15,4))
    cbar = False
    no_ylabel = False
    for i in range(3):
        if i==2:
            cbar=True
        if i>0:
            no_ylabel=True
        image_1mum.plot(i, ax=axes[i], vmax=1e-15, colorbar=cbar, no_ylabel=no_ylabel)
    for ax in axes.ravel():
        ax.set_axis_off()
    fig.savefig('%s/test_data_1_0_scattered_light.png' % path)

def data_1_0_polarisation(path):
    # Plot figure test_data_1_0_polarisation
    image_1mum = mcfost.Image(data_dir + "/data_1.0/")
    fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(15,4))
    cbar = False
    no_ylabel = False
    for i in range(3):
        if i>0:
            no_ylabel=True
        image_1mum.plot(i, ax=axes[i], type="Qphi", vmax=1e-15, colorbar=cbar,
                        no_ylabel=no_ylabel, pola_vector=True, nbin=15)
    for ax in axes.ravel():
        ax.set_axis_off()
    fig.savefig('%s/test_data_1_0_polarisation.png' % path)

def data_1300_scattered_light(path):
    # Plot figure test_data_1300_scattered_light
    image_1mm  = mcfost.Image(data_dir + "/data_1300/")
    fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(15,4))
    cbar = False
    no_ylabel = False
    for i in range(3):
        if i==2:
            cbar=True
        if i>0:
            no_ylabel=True
        image_1mm.plot(i, ax=axes[i], Tb=True, colorbar=cbar, no_ylabel=no_ylabel, vmax=30)
    for ax in axes.ravel():
        ax.set_axis_off()
    fig.savefig('%s/test_data_1300_scattered_light.png' % path)

def data_CO_plot_line(path):
    # Plot figure test_data_CO_plot_line
    mol = mcfost.Line(data_dir + "/data_CO/")
    fig, ax = plt.subplots()
    mol.plot_line(2)
    ax.axis('off')
    fig.savefig('%s/test_data_CO_plot_line.png' % path)

def data_CO_plot_map1(path):
    # Plot figure test_data_CO_plot_map1
    mol = mcfost.Line(data_dir + "/data_CO/")
    fig, ax = plt.subplots()
    mol.plot_map(2, v=0.5, Tb=True, no_xlabel=True, no_ylabel=True, no_xticks=True, no_yticks=True)
    fig.savefig('%s/test_data_CO_plot_map1.png' % path)

def data_CO_plot_map2(path):
    # Plot figure test_data_CO_plot_map2
    mol = mcfost.Line(data_dir + "/data_CO/")
    fig, ax = plt.subplots()
    mol.plot_map(2,v=0.5, bmaj=0.1, bmin=0.1, bpa=0, Tb=True, no_xlabel=True, no_ylabel=True, no_xticks=True, no_yticks=True)
    fig.savefig('%s/test_data_CO_plot_map2.png' % path)


plot_functions = [
    data_th_plot_T,
    data_th_plot_T_log,
    data_th_plot_0,
    data_th_plot_0_contrib,
    data_1_0_scattered_light,
    data_1_0_polarisation,
    data_1300_scattered_light,
    data_CO_plot_line,
    data_CO_plot_map1,
    data_CO_plot_map2,
    ]

@pytest.mark.parametrize("plot_func", plot_functions)
def test_compare_data_th_plot_T(plot_func):
    plot_func(out_path)
    print('./%s/test_%s.png' % (out_path, plot_func.__name__))
    result = compare_images('%s/test_%s.png' % (expected, plot_func.__name__),
                            '%s/test_%s.png' % (out_path, plot_func.__name__),
                            tol=1e-2)
    assert result is None, f"Images not equal: {result}"
